from copy import deepcopy
from typing import Iterable

import numpy as np

from naludaq.controllers import get_peripherals_controller
from naludaq.helpers.helper_functions import extract_event_creation_times, type_name

_missing = object()


class Metadata:
    """Proxy class for accessing metadata stored in a dictionary.

    Metadata in the dictionary is stored according to the following structure::

        {
            'params': {}, # board params dict
            'registers': {}, # board registers dict
            'sensors': [
                {...}, # generated from first call to store_sensor_readings()
                {...}, # generated from second call to store_sensor_readings()
                ...
            ],
            'times': [
                np.ndarray([...]), # generated from first call to store_event_times()
                np.ndarray([...]), # generated from second call to store_event_times()
                ...
            ],
        }
    """

    def __init__(self, metadata: dict = _missing):
        """Create helper proxy class for accessing metadata stored in a dict.

        Args:
            metadata (dict): the destination for metadata. If not provided, a new dict is used.
        """
        self.output = {} if metadata is _missing else metadata

    @property
    def output(self) -> dict:
        """Get/set the output dict for metadata"""
        return self._output

    @output.setter
    def output(self, d: dict):
        if not isinstance(d, dict):
            raise TypeError(f"Metadata output must be dict, not {type_name(d)}")
        self._output = d

    @property
    def params(self) -> dict:
        """Get/set the board params dict"""
        return self.output.setdefault("params", {})

    @params.setter
    def params(self, params: dict):
        if not isinstance(params, dict):
            raise TypeError("Params must be a dict")
        self.output["params"] = deepcopy(params)

    @property
    def registers(self):
        """Get/set the board registers dict"""
        return self.output.setdefault("registers", {})

    @registers.setter
    def registers(self, registers: dict):
        if not isinstance(registers, dict):
            raise TypeError("Registers must be a dict")
        self.output["registers"] = deepcopy(registers)

    @property
    def sensors(self) -> list[dict]:
        """Get the sensor readings list"""
        return self.output.setdefault("sensors", [])

    @property
    def times(self) -> list[np.datetime64]:
        """Get the event creation time list"""
        return self.output.setdefault("times", [])

    def set_configuration(self, board):
        """Convenience method for setting registers/params.

        Args:
            board (Board): a board-like object (i.e. having
                both ``params`` and ``registers`` attributes)
        """
        params = getattr(board, "params", _missing)
        registers = getattr(board, "registers", _missing)
        if params is _missing or registers is _missing:
            raise TypeError("Object must have params and registers attributes")
        self.params = params
        self.registers = registers

    def store_sensor_readings(self, board):
        """Read and store sensor readings into the ``sensors`` list.
        See the class docstring for details on the data structure.

        Args:
            board (Board): connected board object to read sensors from.
            key_format (str): format string to use to generate a unique key for the item.
                The format string must contain a single argument which takes an integer.
        """
        try:
            readings = get_peripherals_controller(board).read_all()
        except Exception:
            raise
        self.sensors.append(readings)

    def store_event_times(self, events: Iterable[dict]) -> np.ndarray:
        """Store event creation times into the ``times`` list.
        See the class docstring for details on the data structure.

        Args:
            events (Iterable[dict]): list of events to store creation times for.

        Returns:
            numpy.ndarray: the stored array.
        """
        self.times.append(extract_event_creation_times(events))



"""
"""
import logging
from collections import defaultdict

from naludaq.communication import DigitalRegisters
from naludaq.helpers.exceptions import InvalidTriggerMaskError
from naludaq.helpers.helper_functions import type_name

from .default import TriggerController

LOGGER = logging.getLogger("naludaq.trigger_controller_upac96")


class TriggerControllerUpac96(TriggerController):
    """TriggerController for UPAC96.

    Based on the default Controller

    """

    def __init__(self, board):
        super().__init__(board)
        self._num_channels = board.params.get("channels", 96)
        self._num_chips = board.params.get("num_chips", 6)
        self._channels_per_chip = self._num_channels // self._num_chips

    @property
    def values(self) -> dict[int, int]:
        """Trigger threshold for the board.

        Set the trigger threshold values, can set one value for all channels or
        a specific value for each channel.

        Updates the value on the hardware if connected.

        Args:
            trigger_val (list or int): Threshold values for triggers per channel.
                Can take an single number for all channels or one value per channel.

        Returns:
            List of current values for the triggers.

        Raises:
            TypeError if not a List[int].
            ValueError if the list is the wrong length
        """
        return self._trigger_values

    @values.setter
    def values(self, trigger_values: "int | list[int] | int") -> None:
        trigger_values = self._convert_trigger_values(trigger_values)
        self.write_triggers(trigger_values=trigger_values)
        self._trigger_values = {c: v for c, v in trigger_values.items()}

    @property
    def enable_trigger_out(self):
        return self.board.registers["control_registers"].get(
            "coincidence_trigger_select", False
        )

    @enable_trigger_out.setter
    def enable_trigger_out(self, value):
        """enable trigger out, enables the trigger output on the board.

        Args:
            value (bool): True for box checked False for box not checked
        Error:
        Raises TypeError if value is not bool
        """
        self._validate_enable_trig_out(value)
        LOGGER.debug("Enable trigger out %s", value)
        self._write_control_register("coincidence_trigger_select", value)

    def get_channel_trigger_mask(self) -> list[int]:
        """Gets the mask for self triggering on individual channels.

        Returns:
            list: a list of enabled channels
        """
        trigger_mask = self.board.params.get("trigger", {}).get("enabled_channels", [])
        return trigger_mask

    def set_channel_trigger_mask(self, channels: list):
        """Sets the mask for self triggering on individual channels.

        The channels which are triggered on will be the ones contained in the list,
        and no other channels will be triggered on.

        Internally this goes chip-by-chip and writes the triggerchannelmask_{lo/hi}
        registers, which control which channels are enabled for self triggering. Any
        chips which have no enabled channels are disabled in the chip trigger mask.

        Args
            channels (list): a list of enabled channels

        Raises:
            TypeError if the given list is not actually a list
            InvalidTriggerMaskError if the list supplied is the wrong length
                or contains invalid elements
        """
        self._validate_channel_list_or_raise(channels)
        for chip in range(self._num_chips):
            chip_channels = self._filter_channels_for_chip(channels, chip)
            trigmask = self._create_trigger_mask(chip_channels)
            self._write_digital_register(
                "triggerchannelmask_lo", trigmask & 0xFF, chips=chip
            )
            self._write_digital_register(
                "triggerchannelmask_hi", trigmask >> 8, chips=chip
            )

        # coincidence trigger on chips without any enabled channels doesn't work
        chip_mask = self._make_chip_mask_from_channel_mask(channels)
        self.set_chip_trigger_mask(chip_mask)
        self.board.params.get("trigger", {})["enabled_channels"] = channels

    def set_single_channel_trigger_mask(self, channel: int):
        """Sets the mask for self triggering on a single channel.

        Args:
            channel (int): the channel to trigger on

        Raises:
            TypeError if the given channel is not an int
            InvalidTriggerMaskError if the channel is out of bounds
        """
        self._validate_channel_list_or_raise([channel])

        self.set_channel_trigger_mask([channel])

    def write_triggers(
        self, trigger_values: "dict[int, int] | int | list[int] | None" = None
    ):
        """Write the trigger values to the hardware.

        This writes the current set triggers to the hardware.
        By setting the trigger_offsets or the trigger_values
        this function is called automatically and the hardware is updated.

        Reimplemented to disable setting wbias/offsets.
        """
        trigger_val = self._convert_trigger_values(trigger_values)
        self._set_trigger_thresholds(trigger_values=trigger_val)

    def _convert_trigger_values(
        self, trigger_values: "int | dict[int, int] | list[int]"
    ) -> dict[int, int]:
        """Convert trigger values to a dict[int, int].

        Args:
            trigger_values: the trigger values to convert

        Returns:
            dict[int, int]: the converted trigger values
        """
        if isinstance(trigger_values, int):
            trigger_val = {c: trigger_values for c in range(self.board.channels)}
        elif isinstance(trigger_values, dict):
            trigger_val = trigger_values
        elif isinstance(trigger_values, list):
            trigger_val = {c: v for c, v in enumerate(trigger_values)}
        else:
            raise TypeError("Trigger values must be a list of integers.")
        return trigger_val

    def compare_channels(self, channels: list[int], prev_channels: list[int]) -> dict:
        """Compare two sets of channels and return the differences.

        Args:
            channels:
            prev_channels:

        Channels to deactivate: present in old but not in new
        Channels to enable: present in new but not in old
        Channels to not touch: present in both old and new
        """
        new = set(channels)
        old = set(prev_channels)
        deactivate = list(old - new)
        enable = list(new - old)
        not_touch = list(new & old)

        return {"deactivate": deactivate, "enable": enable, "leave": not_touch}

    def compare_values(
        self, channels: dict[int, int], prev_channels: dict[int, int]
    ) -> dict[int, int]:
        """Compares the values and creates a new dict containing the channels, values that needs a change.

        Args:
            channels (int): new set of channels to compare against
            prev_channels (int): previous programmed channels

        Returns:
            dict[int, int] of the channels, values that needs to be changed.
        """
        all_chans = set(channels.keys())
        all_chans.update(set(prev_channels.keys()))
        output = {}
        for key in all_chans:
            newval = channels.get(key, None)
            oldval = prev_channels.get(key, None)
            if newval is None:
                pass
            elif oldval != newval:
                output[key] = newval
        return output

    def _make_chip_mask_from_channel_mask(self, channels: list[int]) -> list[int]:
        """Creates a chip trigger mask from the given channel trigger mask.

        Args:
            channels (list[int]): a list of channels to be triggered on

        Returns:
            list[int]: the corresponding list of channels to trigger on
        """
        chip_mask = []
        for chip in range(self._num_chips):
            chip_channels = self._filter_channels_for_chip(channels, chip)
            if len(chip_channels) != 0:
                chip_mask.append(chip)
        return chip_mask

    def _create_trigger_mask(self, chip_channels: list[int]) -> int:
        """Creates a trigger mask for the given channels (single chip only).

        Args:
            chip_channels (list): a list of chip channels to trigger on

        Returns:
            int: the trigger mask for the given channels
        """
        trigmask = sum(1 << (c % self._channels_per_chip) for c in chip_channels)
        trigmask ^= 0xFFFF  # the mask registers are negative logic
        return trigmask

    def _set_trigger_thresholds(
        self, trigger_values: "dict[int, int] | None" = None
    ) -> None:
        """Sets the trigger thresholds from board.trigger_values for all channels"""
        if trigger_values is None:
            trigger_val = self.values
        else:
            self._validate_trigger_dict_or_raise(trigger_values)
            prev_channels = self.values
            trigger_val = self.compare_values(trigger_values, prev_channels)

        threshold_groups = self._group_by_chips(trigger_val)
        for chip, threshold_pair in threshold_groups.items():
            self._write_threshold_pairs(threshold_pair, chip)
        self._reset_writemask()

    def _write_threshold_pairs(self, threshold_pair: dict[int, list[int]], chip: int):
        """Writes the threshold for the given pairs of threshold values and channels.

        Args:
            threshold_pair (dict[int, list[int]]): a dict mapping channels to threshold values.
        """
        for threshold, channels in threshold_pair.items():
            writemask = sum(
                1 << channel % self._channels_per_chip for channel in channels
            )
            self._write_digital_register("writemask_l", writemask & 0xFF, chips=chip)
            self._write_digital_register("writemask_r", writemask >> 8, chips=chip)
            self._write_analog_register("trgthresh", threshold, chips=chip)

    def get_coincidence_trigger_enabled(self) -> "dict[int, bool] | None":
        """Get whether the coincidence trigger is enabled for each chip.

        Returns:
            dict: a dict with chip num (int) as a key and
                coincidence enable as the value (bool)
        """
        return self.board.params.get("trigger", {}).get(
            "coincidence_enabled", {c: False for c in range(self._num_chips)}
        )

    def set_coincidence_trigger_enabled(self, chips: dict[int, bool]):
        """Sets whether the coincidence trigger is enabled for the given chips.

        The coincidence trigger performs "AND"-like logic between channels,
        only emitting a trigger event if multiple channels are triggered at once.
        The period of time for which trigger events are accepted appears to be
        controlled by the wbias register (not set by this function).

        If the coincidence trigger is disabled for any chip, the default "OR"-like
        behavior is performed by the ASIC instead.

        Args:
            chips: a dict with chips (int) as keys, and enabled (bool) for the value
        """
        self._validate_coincidence_trigger_chips_or_raise(chips)
        self.board.trigger.params["coincidence_enabled"] = chips
        for chip, enabled in chips.items():
            # lowest bit of misc reg enables/disables the coincidence trigger
            # for the entire ASIC
            miscreg = self._digital_read("miscreg", chip)
            miscreg = (miscreg & 0b10) | enabled
            self._write_digital_register("miscreg", miscreg, chips=chip)

    def set_chip_trigger_mask(self, chips: list[int]):
        """Sets the chip trigger mask.

        Args:
            chips (list): a list of chips to trigger on

        Raises:
            TypeError if the given list is not actually a list
            InvalidTriggerMaskError if the list supplied is the wrong length
                or contains invalid elements
        """
        LOGGER.debug("Setting chip trigger mask: %s", chips)
        chipmask = sum(1 << chip for chip in chips)
        self._write_control_register("udc_trigger_mask", chipmask)

    def set_trigger_edge(self, channel_edges: dict[int, bool]):
        """Set which signal edge to trigger on per channel.

        Shift between positive going signals (rising) and negative (falling).
        The ASIC will not trigger on the signal until it detects the appropriate
        edge.

        Args:
            channel_edges (list[bool]): list of edge settings for each channel.
                Each value should be True for rising edges, or False for falling edges.
        """
        channel_edges = [channel_edges.get(c, False) for c in range(self._num_channels)]

        self._validate_edge_list_or_raise(channel_edges)
        LOGGER.debug("Setting trigger edges: %s", channel_edges)
        for chip in range(self._num_chips):
            edge_mask = self._extract_edge_mask_for_chip(channel_edges, chip)
            self._set_sgn(chip, edge_mask)
        self._reset_writemask()

    def _set_sgn(self, chip: int, edge_mask: int):
        """Set the sgn register.

        Highest bit is broadcast bit, it's required to be set
        in order to write to each copy of the sgn register individually

        Args:
            edge_mask (int): the edge mask for the given chip
        """
        sgn_l, sgn_r = self._flip_sgn_bytes(edge_mask)
        self._write_digital_register("writemask_l", 0x1FF, chips=chip)
        self._write_digital_register("writemask_r", 0, chips=chip)
        self._write_analog_register("sgn", sgn_l, chips=chip)
        self._write_digital_register("writemask_l", 0, chips=chip)
        self._write_digital_register("writemask_r", 0x1FF, chips=chip)
        self._write_analog_register("sgn", sgn_r, chips=chip)

    def _extract_edge_mask_for_chip(self, channel_edges: list[bool], chip: int) -> int:
        """Extract the edge mask for the given chip from the given list of edges.

        Args:
            channel_edges (list[bool]): list of edge settings for each channel.
                Each value should be True for rising edges, or False for falling edges.
            chip (int): the chip to extract the edge mask for

        Returns:
            int: the edge mask for the given chip
        """
        return sum(
            1 << c
            for c in range(self._channels_per_chip)
            if channel_edges[c + chip * self._channels_per_chip]
        )

    def _flip_sgn_bytes(self, sgn_mask: int) -> tuple[int, int]:
        """Flip the bytes of the sgn register mask.

        Individual bytes need to be flipped because the bits in
        the sgn register are reversed (i.e. bit 0 corresponds to channel 7).

        Args:
            sgn_mask (int): the full sgn mask. MSB corresponds to the right
                side, and lower bytes correspond to the left side.

        Returns:
            tuple[int, int]: sgn for the left side, and sgn for the right side.
        """

        def flip_byte(n):
            return int(f"{n:08b}"[::-1], 2)

        sgn_l = flip_byte(sgn_mask & 0xFF)
        sgn_r = flip_byte(sgn_mask >> 8)
        return sgn_l, sgn_r

    def _group_by_chips(self, pairs: dict[int, int]) -> dict[int, dict[int, list[int]]]:
        """Groups a list into a dict where the key is the chip,
        the value is a dict(list) which maps all unique values
        to the channels that correspond to them.

        Example::

            {
                0: { # chip
                    1000: [0, 1, 2, 3] # threshold: channels
                }
            }
        ```
        """
        groups = defaultdict(lambda: defaultdict(list))
        for channel, val in pairs.items():
            chip = channel // self._channels_per_chip
            groups[chip][val].append(channel)
        return groups

    def _filter_channels_for_chip(self, channels: list[int], chip: int) -> list[int]:
        """Filters out channels in the given list that don't belong to the given chip."""
        return list(filter(lambda c: (c // self._channels_per_chip) == chip, channels))

    def _reset_writemask(self):
        """Reset the write mask register for all chips.

        The writemask register is restored to 0x1FF (broadcast + all channels),
        regardless of what it was set to before. This is important because
        it ensures the other analog registers don't get written improperly
        elsewhere due to the writemask being wrong.
        """
        self._write_digital_register("writemask_l", 0x1FF)
        self._write_digital_register("writemask_r", 0x1FF)

    def _digital_read(self, name: str, chip: int) -> int:
        """Read the value of a digital register"""
        return DigitalRegisters(self.board, chips=chip).read(name)["value"]

    def _validate_channel_list_or_raise(self, channels):
        """Validates a list of channels, and raises an error
        if there's a problem with them.

        Args:
            channels: the object that is supposedly a channels list

        Raises:
            TypeError if the channel list is not a list
            InvalidTriggerMaskError if the list is too long,
                contains elements that are not an int,
                or contains elements that are too large.
        """
        if not isinstance(channels, list):
            raise TypeError(f"Channels needs to be a list, not a {type(channels)}")
        if len(channels) > self._num_channels:
            raise InvalidTriggerMaskError(
                f"Too many channels, max is {self._num_channels}"
            )
        for channel in channels:
            if not isinstance(channel, int):
                raise InvalidTriggerMaskError("Channels can only contain integers")
            if channel >= self._num_channels or channel < 0:
                raise InvalidTriggerMaskError(f"Channel {channel} is out of bounds")

    def _validate_trigger_list_or_raise(self, trigger_list: list):
        """Validates a list of thresholds, and raises an error
        if there's a problem with them.

        Args:
            trigger_list: a list containing thresholds values per channel

        Raises:
            TypeError if the trigger_list is not a list
            ValueError if theres more list values than channels,
                contains elements that are not an int,
                or contains elements that are too large.
        """
        if not isinstance(trigger_list, list):
            raise TypeError(
                f"trigger_list needs to be a list, not a {type_name(trigger_list)}"
            )
        if len(trigger_list) != self._num_channels:
            raise ValueError(
                f"trigger_list length: {len(trigger_list)} needs to match num channels: {self._num_channels}"
            )
        for threshold in trigger_list:
            if not isinstance(threshold, int):
                raise ValueError("Thresholds can only contain integers")
            if threshold > self._max_thresholds or threshold < self._min_thresholds:
                raise ValueError(f"Threshold {threshold} is out of bounds")

    def _validate_trigger_dict_or_raise(self, trigger_dict: dict[int, int]):
        """Validates a dictionary of thresholds, and raises an error
        if there's a problem with them.

        Args:
            trigger_dict: a dict containing thresholds values per channel

        Raises:
            TypeError if the trigger_dict is not a dict
            ValueError if theres more dict values than channels,
                contains elements that are not an int,
                or contains elements that are too large.
        """
        if not isinstance(trigger_dict, dict):
            raise TypeError(
                f"trigger_dict needs to be a dict, not a {type_name(trigger_dict)}"
            )
        if len(trigger_dict) > self._num_channels:
            raise ValueError(
                f"trigger_dict length: {len(trigger_dict)} needs to match num channels: {self._num_channels}"
            )
        for channel, threshold in trigger_dict.items():
            if not isinstance(channel, int):
                raise ValueError(f"Channel {channel} is not an int")
            if not isinstance(threshold, int):
                raise ValueError(f"Threshold {threshold} is not an int")
            if not self._min_thresholds <= threshold <= self._max_thresholds:
                raise ValueError(
                    f"Threshold {threshold} is out of bounds ({self._min_thresholds}:{self._max_thresholds+1})"
                )

    def _validate_edge_list_or_raise(self, edge_list: list):
        """Validates a list of edges of bools, and raises an error
        if there's a problem with them.

        Args:
            edge_list: a list containing thresholds values per channel

        Raises:
            TypeError if the edge_list is not a list
            ValueError if theres more list values than channels,
                contains elements that are not a bool,
                or contains elements that are too large.
        """
        if not isinstance(edge_list, list):
            raise TypeError(
                f"trigger_list needs to be a list, not a {type_name(edge_list)}"
            )
        if len(edge_list) != self._num_channels:
            raise ValueError(
                "edge_list is not the same length as the number of channels"
            )
        if any(not isinstance(edge, bool) for edge in edge_list):
            raise ValueError("edge_list can only contain bools")

    def _validate_coincidence_trigger_chips_or_raise(self, chips: dict):
        """Validate whether the arguments to the coincidence trigger setter are valid

        Args:
            chips (dict): a dict with chip num (int) as a key and
                coincidence enable as the value (bool)

        Raises:
            TypeError: if chips is not a dict
            ValueError: if chips keys contains chips that are out of range of num chips
            TypeError: if chips values contains elements that are not a bool
        """
        if not isinstance(chips, dict):
            raise TypeError(f"chips needs to be a dict, not a {type_name(chips)}")
        for k, v in chips.items():
            if k not in range(self._num_chips):
                raise ValueError(f"chips contains an invalid key: {k}")
            if not isinstance(v, bool):
                raise TypeError(f"chip contains an invalid value: {v}")

    def _validate_enable_trig_out(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"value needs to be a bool, not a {type_name(value)}")



"""BoardInits is a Factory for the init_board command lists.

This class returns a function that should initialize each board,
since they are all different.

"""
import logging
import time

from naludaq.backend.managers.connection import ConnectionManager
from naludaq.board.initializers import Initializers
from naludaq.board.initializers.aardvarcv3 import InitAardvarcv3
from naludaq.board.initializers.aodsoc import InitAodsoc
from naludaq.board.initializers.asocv3s import InitASoCv3S
from naludaq.board.initializers.hiper import InitHiper
from naludaq.board.initializers.init_aodsv2_eval import InitAodsv2Eval
from naludaq.board.initializers.init_hdsocv1 import InitHDSoCv1
from naludaq.board.initializers.init_hdsocv2 import InitHDSoCv2
from naludaq.board.initializers.init_udc import InitUDC16
from naludaq.board.initializers.init_upac96 import InitUPAC96
from naludaq.board.initializers.trbhm import InitTrbhm
from naludaq.communication import (
    AnalogRegisters,
    ControlRegisters,
    DigitalRegisters,
    I2CRegisters,
)
from naludaq.communication.i2c import readI2cReg, sendI2cCommand

LOGGER = logging.getLogger("naludaq.board_inits")


class BoardInits:
    """Factory for the init_board commands.
    Hardware paramters for the chip/board and
    changes when hardware changes.
    By having this class the actual storage of the paramters are
    removed from the logic of the rest o the program.

    Functions:
        init_board()

    """

    def __init__(self, board):
        self.board = board
        self._initializers = {
            "aardvarcv2": self.init_AARDVARCv2,
            "aardvarcv3": InitAardvarcv3,
            "aardvarcv4": InitAardvarcv3,
            "aodsv1": self.init_AODSv1,
            "aodsv2_eval": InitAodsv2Eval,
            "asoc": self.init_ASoC,
            "asocv2": self.init_ASoCv2,
            "asocv3": self.init_ASoCv3,
            "asocv3s": InitASoCv3S,
            "hdsocv1": InitHDSoCv1,
            "hdsocv1_evalr1": InitHDSoCv1,
            "hdsocv1_evalr2": InitHDSoCv1,
            "hdsocv2_eval": InitHDSoCv2,
            "hdsocv2_evalr2": InitHDSoCv2,
            "hiper": InitHiper,
            "aodsoc_asoc": InitAodsoc,
            "aodsoc_aods": InitAodsoc,
            "siread": self.init_SiREAD,
            "trbhm": InitTrbhm,
            "dsa-c10-8": InitTrbhm,
            "udc16": InitUDC16,
            "upac96": InitUPAC96,
            "upac32": self.init_upac32,
        }

    def run_init_function(self) -> bool:
        """Runs the initialization sequence for the board.

        Returns:
            True if the board was successfully initialized, False otherwise.
        """
        model = self.board.params["model"]
        LOGGER.info("Init board - %s", model)

        initializer = self._initializers.get(model, None)
        if initializer is None:
            LOGGER.error("Error in board_inits: model %s not found", model)
            return False
        elif isinstance(initializer, type) and issubclass(initializer, Initializers):
            return initializer(self.board).run()
        else:
            return initializer()

    def init_ASoC(self) -> bool:
        """Initialize the board functions.

        Run through the startup script to init the board.

        Returns:
            True if the board is successfully connected, False if connection failed.
        """
        ControlRegisters(self.board).write_all()

        ControlRegisters(self.board).write("3v3_en_main", True)
        time.sleep(0.25)
        LOGGER.info("Programming clock")
        self.board.clock.program(filename=self.board.params["default_si5341"])
        time.sleep(0.25)
        # Starting the clock output, or no clocks signal on the board.
        ControlRegisters(self.board).write("clk_oe", True)
        ControlRegisters(self.board).write("sysrst", False)
        self.analogStartup()
        self.board.ext_bias.set_dacs(self.board.params["ext_dac"])

        sendI2cCommand(self.board, "30", ["05"])  # tempSensor to temp reg
        sendI2cCommand(self.board, "D0", ["00111011"])  # set up current sense ADC
        self.board.trigger.values = self.board.trigger.params.get(
            "default_trigger_value", 1350
        )
        self.board.write_triggers()
        LOGGER.info("Programming FPGA settings")
        DigitalRegisters(self.board).write_all()
        ControlRegisters(self.board).write("ls_oe", True)
        self.board.connection.DEBUG = False

        return True

    def init_ASoCv2(self) -> bool:
        """Initialize the board functions.

        Run through the startup script to init the board.

        Returns:
            True if the board is successfully connected, False if connection failed.
        """
        ControlRegisters(self.board).write_all()
        self._i2c_startup()

        ControlRegisters(self.board).write("2v5_en", True)
        time.sleep(0.25)
        LOGGER.info("Programming clock")
        self.board.clock.program(filename=self.board.params["clock_file"])
        time.sleep(0.25)
        ControlRegisters(self.board).write("clk_noe", False)
        ControlRegisters(self.board).write("sysrst", True)
        ControlRegisters(self.board).write("sysrst", False)
        ControlRegisters(self.board).write("iomode0", True)
        ControlRegisters(self.board).write("iomode1", False)

        self.analogStartup()
        DigitalRegisters(self.board).write_all()
        self.board.ext_bias.set_dacs(self.board.params["ext_dac"])

        I2CRegisters(self.board).write_all()
        sendI2cCommand(self.board, "30", ["05"])  # tempSensor to temp reg
        sendI2cCommand(self.board, "D0", ["00111011"])  # set up current sense ADC

        LOGGER.info("Programming FPGA settings")

        return True

    def init_ASoCv3(self) -> bool:
        """Initialize the board functions.

        Run through the startup script to init the board.

        Returns:
            True if the board is successfully connected, False if connection failed.
        """
        ControlRegisters(self.board).write_all()
        self._i2c_startup()

        ControlRegisters(self.board).write("2v5_en", True)
        time.sleep(0.25)
        LOGGER.info("Programming clock")
        self.board.clock.program(filename=self.board.params["clock_file"])
        time.sleep(0.25)
        ControlRegisters(self.board).write("clk_noe", False)
        ControlRegisters(self.board).write("sysrst", True)
        ControlRegisters(self.board).write("sysrst", False)
        ControlRegisters(self.board).write("iomode0", True)
        ControlRegisters(self.board).write("iomode1", False)

        self.analogStartup()
        DigitalRegisters(self.board).write_all()
        ext_dac_val = self.board.params.get("ext_dac", {})
        for chan, dac_val in ext_dac_val["channels"].items():
            self.board.ext_bias.set_single_dac(
                chan, dac_val
            )  # self.board.params['ext_dac'])

        sendI2cCommand(self.board, "30", ["05"])  # tempSensor to temp reg
        sendI2cCommand(self.board, "D0", ["00111011"])  # set up current sense ADC

        LOGGER.info("Programming FPGA settings")

        return True

    def init_AODSv1(self) -> bool:
        """Initialize the board functions.

        Run through the startup script to init the board.

        Returns:
            True if the board is successfully connected, False if connection failed.
        """
        ControlRegisters(self.board).write_all()
        self._i2c_startup()

        ControlRegisters(self.board).write("2v5_en", True)
        time.sleep(0.25)
        LOGGER.info("Programming clock")
        self.board.clock.program(filename=self.board.params["clock_file"])
        time.sleep(0.25)
        ControlRegisters(self.board).write("clk_noe", False)
        ControlRegisters(self.board).write("sysrst", True)
        ControlRegisters(self.board).write("sysrst", False)
        ControlRegisters(self.board).write("iomode0", True)
        ControlRegisters(self.board).write("iomode1", False)

        self.analogStartup()
        DigitalRegisters(self.board).write_all()
        ext_dac_val = self.board.params.get("ext_dac", {})
        for chan, dac_val in ext_dac_val["channels"].items():
            self.board.ext_bias.set_single_dac(
                chan, dac_val
            )  # self.board.params['ext_dac'])

        I2CRegisters(self.board).write_all()
        sendI2cCommand(self.board, "30", ["05"])  # tempSensor to temp reg
        sendI2cCommand(self.board, "D0", ["00111011"])  # set up current sense ADC

        # AODS specific
        DigitalRegisters(self.board).write("waitread", 1)

        LOGGER.info("Programming FPGA settings")

        return True

    def init_SiREAD(self) -> bool:
        ControlRegisters(self.board).write_all()
        self._i2c_startup()

        time.sleep(0.1)
        ControlRegisters(self.board).write("v2v5_en", True)

        self.change_si570(100)

        AnalogRegisters(self.board).write_all()

        time.sleep(0.1)
        for _ in range(5):
            ControlRegisters(self.board).write("wraddrsync", True)
            time.sleep(0.1)
            ControlRegisters(self.board).write("wraddrsync", False)
            time.sleep(0.1)
        time.sleep(0.1)

        sendI2cCommand(self.board, "30", ["05"])  # tempSensor to temp reg
        sendI2cCommand(self.board, "D0", ["00111011"])  # set up current sense ADC

        return True

    def _power_toggle(self, state):
        for register in [
            "2v5_en",
            "1v2_en",
            "3v3_i2c_en",
            "clk2v5_en",
            "clk1v8_en",
            "clk_i2c_sel",
        ]:
            ControlRegisters(self.board).write(register, state)

    def init_AARDVARCv2(self) -> bool:
        ControlRegisters(self.board).write_all()
        self._i2c_startup()

        self._power_toggle(False)
        self._power_toggle(True)

        time.sleep(0.25)
        LOGGER.info("Programming clock")
        self.board.clock.program(filename=self.board.params["clock_file"])
        time.sleep(0.25)
        ControlRegisters(self.board).write("clk_oeb", False)
        ControlRegisters(self.board).write("sysrst", True)
        ControlRegisters(self.board).write("sysrst", False)
        ControlRegisters(self.board).write("iomode0", True)
        ControlRegisters(self.board).write("iomode1", False)

        self.analogStartup()
        DigitalRegisters(self.board).write_all()
        ext_dac_val = self.board.params.get("ext_dac", {})
        for chan, dac_val in ext_dac_val["channels"].items():
            self.board.ext_bias.set_single_dac(chan, dac_val)

        I2CRegisters(self.board).write_all()
        sendI2cCommand(self.board, "30", ["05"])  # tempSensor to temp reg
        sendI2cCommand(self.board, "D0", ["00111011"])  # set up current sense ADC

        return True

    def init_HIPeR(self) -> bool:
        ControlRegisters(self.board).write_all()
        self._i2c_startup()

        self._power_toggle(False)
        self._power_toggle(True)

        time.sleep(0.25)
        LOGGER.info("Programming clock")
        self.board.clock.program(filename=self.board.params["clock_file"])
        time.sleep(0.25)
        ControlRegisters(self.board).write("clk_oeb", False)
        ControlRegisters(self.board).write("sysrst", True)
        ControlRegisters(self.board).write("sysrst", False)
        ControlRegisters(self.board).write("iomode0", True)
        ControlRegisters(self.board).write("iomode1", False)

        self.analogStartup()
        DigitalRegisters(self.board).write_all()
        ext_dac_val = self.board.params.get("ext_dac", {})
        for chan, dac_val in ext_dac_val["channels"].items():
            self.board.ext_bias.set_single_dac(
                chan, dac_val
            )  # self.board.params['ext_dac'])

        I2CRegisters(self.board).write_all()
        sendI2cCommand(self.board, "30", ["05"])  # tempSensor to temp reg
        sendI2cCommand(self.board, "D0", ["00111011"])  # set up current sense ADC

    def init_upaci(self) -> bool:
        """Starting the UPAC-I board."""
        if self.board.using_new_backend:
            return self._init_upac32_new()
        # ControlRegisters(self.board).write_all_registers()
        LOGGER.info("Setting up UPAC-I")
        if (
            self.board.connection.baud >= 1_000_000
            or self.board.connection_info["speed"] >= 1_000_000
        ):
            self.board.connection.rtscts = True
            ControlRegisters(self.board).write("uart_handshake_en_out", True)
        else:
            self.board.connection.rtscts = False
            ControlRegisters(self.board).write("uart_handshake_en_out", False)
        try:
            ctrl_regs = ControlRegisters(self.board).read_all()
        except Exception:
            LOGGER.error("Can't read the control registers off the board.")
        else:
            self.board.registers["control_registers"] = ctrl_regs

        dac_val = 30_000
        ControlRegisters(self.board).write("dac_vbias00_07_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias08_15_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias16_23_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias24_31_out", dac_val)
        ControlRegisters(self.board).write("dac_write_strobe", True)
        ControlRegisters(self.board).write("dac_write_strobe", False)

        return True

    def init_upac32(self) -> bool:
        """Starting the upac32 board."""
        if self.board.using_new_backend:
            return self._init_upac32_new()

        # ControlRegisters(self.board).write_all_registers()
        LOGGER.info("Setting up UPAC32")
        # ControlRegisters(self.board).write("baud_rate_divisor", 108)
        if (
            self.board.connection.baud >= 1_000_000
            or self.board.connection_info["speed"] >= 1_000_000
        ):
            self.board.connection.rtscts = True
            ControlRegisters(self.board).write("uart_handshake_en_out", True)
        else:
            self.board.connection.rtscts = False
            ControlRegisters(self.board).write("uart_handshake_en_out", False)

        try:
            ctrl_regs = ControlRegisters(self.board).read_all()
        except Exception:
            LOGGER.error("Can't read the control registers off the board.")
        else:
            self.board.registers["control_registers"] = ctrl_regs

        dac_val = 30_000
        ControlRegisters(self.board).write("dac_vbias00_07_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias08_15_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias16_23_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias24_31_out", dac_val)
        ControlRegisters(self.board).write("dac_write_strobe", True)
        ControlRegisters(self.board).write("dac_write_strobe", False)

        return True

    def _init_upac32_new(self) -> bool:
        """Starting the upac32 board."""
        LOGGER.info("Setting up UPAC32")
        device = ConnectionManager(self.board)
        if (
            device.baud_rate >= 1_000_000
            or self.board.connection_info["baud_rate"] >= 1_000_000
        ):
            device.rts_cts = True
            ControlRegisters(self.board).write("uart_handshake_en_out", True)
        else:
            device.rtscts = False
            ControlRegisters(self.board).write("uart_handshake_en_out", False)

        try:
            ctrl_regs = ControlRegisters(self.board).read_all()
        except Exception:
            LOGGER.error("Can't read the control registers off the board.")
        else:
            self.board.registers["control_registers"] = ctrl_regs

        dac_val = 30_000
        ControlRegisters(self.board).write("dac_vbias00_07_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias08_15_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias16_23_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias24_31_out", dac_val)
        ControlRegisters(self.board).write("dac_write_strobe", True)
        ControlRegisters(self.board).write("dac_write_strobe", False)

        return True

    def init_zdigitizer(self) -> bool:
        """Starting the Z-Digitizer board."""
        if self.board.using_new_backend:
            return self._init_upac32_new()
        # ControlRegisters(self.board).write_all_registers()
        LOGGER.info("Setting up Z-Digitizer")
        if (
            self.board.connection.baud >= 1_000_000
            or self.board.connection_info["speed"] >= 1_000_000
        ):
            self.board.connection.rtscts = True
            ControlRegisters(self.board).write("uart_handshake_en_out", True)
        else:
            self.board.connection.rtscts = False
            ControlRegisters(self.board).write("uart_handshake_en_out", False)
        try:
            ctrl_regs = ControlRegisters(self.board).read_all()
        except Exception:
            LOGGER.error("Can't read the control registers off the board.")
        else:
            self.board.registers["control_registers"] = ctrl_regs

        dac_val = 30_000
        ControlRegisters(self.board).write("dac_vbias00_07_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias08_15_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias16_23_out", dac_val)
        ControlRegisters(self.board).write("dac_vbias24_31_out", dac_val)
        ControlRegisters(self.board).write("dac_write_strobe", True)
        ControlRegisters(self.board).write("dac_write_strobe", False)

        return True

    def change_si570(self, freq):
        """
        change the clock frequency of the si570 by writing to register 13
        Values taken from the XO calculator (start-up freq = 156.25)
        """
        board = self.board
        board.connection.reset_input_buffer()
        board.connection.reset_output_buffer()
        addr = "BA"

        hs_map = [
            (4, "000"),
            (5, "001"),
            (6, "010"),
            (7, "011"),
            (9, "101"),
            (11, "111"),
        ]

        # restart chip
        sendI2cCommand(board, addr, (["87", "01"]))
        time.sleep(2)

        # read start up freq configuration
        temp = readI2cReg(board, addr, "07")
        hs_div_bin = temp >> 5
        hs_div = -1
        for hs_div_i in hs_map:
            if int(hs_div_i[1], 2) == hs_div_bin:
                hs_div = hs_div_i[0]
                break
        if hs_div == -1:
            LOGGER.error("no valid hs_div found... %s", hs_div_bin)
            return

        n1 = (temp & 31) << 2
        temp = readI2cReg(board, addr, "08")
        n1 += temp >> 6
        n1 += 1

        rfreq = (temp & 63) << 32
        rfreq += readI2cReg(board, addr, "09") << 24
        rfreq += readI2cReg(board, addr, "0A") << 16
        rfreq += readI2cReg(board, addr, "0B") << 8
        rfreq += readI2cReg(board, addr, "0C")
        rfreq2 = rfreq / (2**28)

        LOGGER.debug("hs_div=%s , n1=%s", hs_div, n1)
        LOGGER.debug("rfreq=%s , rfreq/(2^28)=%s", hex(rfreq), rfreq2)

        fxtal = (156.25 * hs_div * n1) / rfreq2

        LOGGER.debug("fxtal=%s", fxtal)

        # calculate new values
        good = False

        for hs_div_i in hs_map:
            for n1 in range(0, 128, 2):
                fdco = hs_div_i[0] * n1 * freq
                if fdco > 4850 and fdco < 5670:
                    LOGGER.debug("fdco=%s @n1=%s, hs_div_i=%s", fdco, n1, hs_div_i)
                    good = True
                    hs_div = hs_div_i
                    break
            if good:
                break

        if not good:
            LOGGER.debug("no valid values?")
            return
        LOGGER.debug(hs_div)

        rfreq = fdco / fxtal
        rfreqHex = hex(int(rfreq * 2**28))[2:].zfill(10)
        LOGGER.debug("rfreqHex:%s hs_dev[0]:%s n1:%s", rfreqHex, hs_div[0], n1)

        new_freq1 = hex(int(int(hs_div[1], 2) * (2**5) + (n1 - 1) / (2**2)))[2:].zfill(
            2
        )
        new_freq2 = (
            hex(((n1 - 1) % 4) * (2**6) + int(rfreqHex[0], 16) * (2**5))[2:-1]
            + rfreqHex[1]
        )

        sendI2cCommand(board, addr, ["89", "10"])  # freeze DCO
        sendI2cCommand(board, addr, ["07", new_freq1])
        sendI2cCommand(board, addr, ["08", new_freq2])
        sendI2cCommand(board, addr, ["09", rfreqHex[2:4]])
        sendI2cCommand(board, addr, ["0A", rfreqHex[4:6]])
        sendI2cCommand(board, addr, ["0B", rfreqHex[6:8]])
        sendI2cCommand(board, addr, ["0C", rfreqHex[8:]])
        sendI2cCommand(board, addr, ["89", "00"])  # unfreeze DCO
        sendI2cCommand(board, addr, ["87", "40"])  # set NewFreq bit within 10ms

    def analogStartup(self):
        model = self.board.params["model"]
        if model in ["asoc", "asoc-ml605", "asocv2", "asocv3", "aodsv1"]:
            self.analogStartup_ASoC()

        elif model in ["aardvarcv2"]:
            self.analogStartup_AARDVARC()

        elif model == "siread":
            AnalogRegisters(self.board).write_all()

        else:
            LOGGER.error("Error: analogStartup() didn't recognize model %s", model)

    def analogStartup_ASoC(self):
        """Startup sequence for the analog.

        Same as analog update, except it sets and unsets vanbuff to get the dll going.
        """
        LOGGER.debug("Setting analog ASoC control registers")
        board = self.board
        AnalogRegisters(board).write_all()
        AnalogRegisters(board).write("qbuff", 0)
        AnalogRegisters(board).write("vanbuff", 0xB00)
        time.sleep(1)
        AnalogRegisters(board).write("qbuff", 2048)
        AnalogRegisters(board).write("vanbuff", 0)

    def analogStartup_AARDVARC(self):
        """Startup sequence for the analog.

        Same as analog update, except it sets and unsets vanbuff to get the dll going.
        """
        board = self.board
        LOGGER.debug("Setting analog ASoC control registers")
        AnalogRegisters(board).write_all()
        AnalogRegisters(board).write("qbias", 0)
        AnalogRegisters(board).write("vadjn_sw", True)
        time.sleep(1)
        AnalogRegisters(board).write("qbias", 2048)
        AnalogRegisters(board).write("vadjn_sw", False)

    def _i2c_startup(self):
        """Startup for I2C. Just powers it on."""
        I2CRegisters(self.board).write("i2c_en", True)



import time
from logging import getLogger
from threading import Thread

from naludaq.helpers.exceptions import BadDataError

logger = getLogger(__name__)


class ParserWorkerLight(Thread):
    """Lightweight replacement for the parser worker."""

    def __init__(
        self, parser, input_buffer, output_buffer, error_buffer, frequency=1000
    ):
        super().__init__()
        self.parser = parser
        self.input_buffer = input_buffer
        self.output_buffer = output_buffer
        self.error_buffer = error_buffer
        self.frequency = frequency
        self.running = True
        self.daemon = True

    @property
    def frequency(self):
        """Get/Set the frequency this thread polls the serial port.

        Frequency in Hz (times per second).
        """

        return self._frequency

    @frequency.setter
    def frequency(self, freq):
        if not isinstance(freq, (int, float)):
            raise TypeError("Frequency must be an integer or float")
        if freq <= 0:
            raise ValueError("Frequency must be positive")

        self._frequency = freq
        self._interval = 1 / freq

    def run(self):
        self._parserloop()

    def stop(self):
        self.running = False

    def _parserloop(self):
        """Main worker loop"""
        evt_num = 0
        try:
            while self.running or self.input_buffer:
                frametime = time.perf_counter()
                next_pkg = None

                if self.input_buffer:
                    next_pkg = self.input_buffer.popleft()

                if next_pkg:
                    try:
                        parsed_package = self.parser.parse(next_pkg)
                    except (Exception, TypeError, BadDataError) as error_msg:
                        logger.error("Parsing package failed due to: %s", error_msg)
                        self.error_buffer.append(next_pkg)
                    else:
                        parsed_package["event_num"] = evt_num
                        self.output_buffer.append(parsed_package)
                        evt_num += 1

                try:
                    time.sleep(frametime + self._interval - time.perf_counter())
                except ValueError:
                    pass
        except KeyboardInterrupt:
            pass



from naludaq.parsers.headers.base import HeaderParser


class SiREADHeaderParser(HeaderParser):
    @staticmethod
    def parse_header(event: dict, raw_data: bytes):
        """Parse event headers from raw data and store them into an event dict.

        Args:
            event (dict): the event
            raw_data (bytes): the raw data to parse.
        """
        # event['prev_final_window'] = (raw_data[0]&255) # evt header 0
        trigger_time_ns = raw_data[0] << 12  # 1
        trigger_time_ns += raw_data[1]  # 2 # 24 bit number
        event["trigger_time_ns"] = trigger_time_ns
        event["event_id"] = raw_data[2]  # 3



"""Readout settings for the aardvarcv3 board.

When setting the readout channel, aardvarc uses exclude channel mask registers
to exclude certain channels from readout

Controls the readout channels and the readout windows.
"""
from logging import getLogger

from .default import ReadoutController

LOGGER = getLogger("naludaq.readout_controller_aardvarcv3")


class Aardvarcv3ReadoutController(ReadoutController):
    def __init__(self, board):
        """Readout Controller for aardvarcv3.
        Args:
            board (Board): the board object.
        """
        super().__init__(board)

    def set_readout_channels(self, channels_to_read: list):
        """Select channels to readout.

        Update the registers and write them to the board.

        Args:
            channels_to_read(list): List of channel numbers to read

        Raises:
            TypeError if the list or an element in the list are the wrong type
            ValueError if the list is too large, too small, or contains a channel
                number that is out of bounds.

        """
        self._validate_channels_or_raise(channels_to_read)
        chansel = self._generate_channels_bits(channels_to_read)
        max_chan = self.board.params["channels"]
        # Exclude channel mask is opposite of chansel
        excludechanmask = (2**max_chan - 1) ^ int(chansel, 2)
        self._write_digital_register("excludechannelmask", excludechanmask)

    def get_readout_channels(self):
        """Get the current channels to read out.

        Returns:
            Sorted list of channel numbers to read out.
        """
        mask_reg = self.board.registers["digital_registers"]["excludechannelmask"]
        mask = mask_reg["value"][0]
        return [c for c in range(self.board.channels) if (mask >> c) & 1 == 0]

    def disable_calibration_channel(self, channels: "int | list[int]" = None):
        """Disable use of the calibration channel for all channels."""
        self._set_cal_mode_or_raise(channels, False, False)

    def enable_external_calibration_channel(self, channels: "int | list[int]"):
        """Enable external calibration for the given channels."""
        self._set_cal_mode_or_raise(channels, True, True)

    def enable_internal_calibration_channel(self, channels: "int | list[int]"):
        """Enable internal calibration for the given channels."""
        self._set_cal_mode_or_raise(channels, True, False)

    def _set_cal_mode_or_raise(
        self,
        channels: "int | list[int] | None",
        enabled: bool,
        external: bool,
    ):
        """Set the calibration mode for each of the given channels.
        Channels not included in the given list will not be modified.

        Args:
            channels (int | list[int] | None): the channel(s) to set the calibration for,
                or `None` for all channels.
            enabled (bool): whether to enable calibration for the given channel(s)
            external (bool): whether to use the external calibration channel for the
                given channels
        """
        if channels is None:
            channels = list(range(self.board.channels))
        if not isinstance(enabled, bool) or not isinstance(external, bool):
            raise TypeError("Enabled and external must be booleans")
        if isinstance(channels, int):
            channels = [channels]
        self._validate_channels_or_raise(channels, min_valid=0)

        channels_per_chip = self.board.channels // self.board.available_chips
        for channel in channels:
            chip = channel // channels_per_chip
            rel_channel = channel % channels_per_chip
            self._write_analog_register(f"cal_en_{rel_channel}", enabled, chips=chip)
            self._write_analog_register(f"cal_ext_{rel_channel}", external, chips=chip)
        for chip in range(self.board.available_chips):
            analog = self.board.registers["analog_registers"]
            global_en = any(
                analog[f"cal_en_{c}"]["value"][chip] for c in range(channels_per_chip)
            )
            self._write_analog_register("cal_en", global_en, chips=chip)



"""Legal stuff

"""
import copy
from logging import getLogger

import numpy as np

from naludaq.helpers.exceptions import BadDataError

LOGGER = getLogger(__name__)


class PedestalsCorrecter:
    """Pedestals correct

    Apply the pedestals correction to events and remove it later if necessary.
    Correct an event with previously generated pedestals or remove the correction.
    Removing the correction requires the pedestals to be stored.

    Attributes:
        params: dictionary with channel, windows, and samples
        pedestals

    """

    def __init__(self, params: dict = None, pedestals: dict = None):
        self.params = params
        self._pedestals = None
        self.pedestals = pedestals

    @property
    def pedestals(self):
        """Get/Set the pedestals for the parser to use.

        Can be set to None to run without subtraction.
        Doesn't raise an error if invalid pedestals, just sets the value to None.
        """
        return self._pedestals

    @pedestals.setter
    def pedestals(self, peds):
        if isinstance(peds, type(None)):
            self._pedestals = None
        else:
            shape = (
                self.params["channels"],
                self.params["windows"] * self.params["samples"],
            )
            channels = np.ones(shape=shape)
            for chan in range(self.params["channels"]):
                channels[chan] = np.concatenate(peds["data"][chan]).ravel()[0:]

            self._pedestals = channels

    def run(self, event: dict, correct_in_place: bool = True) -> dict:
        """Pedestals correct an event.

        The function assumes the board is carrying the pedestals,
        if there are no pedestals on the board just returns the event unchanged.

        Args:
            event (dict or object):
            correct_in_place (bool): Whether to directly modify the event
                passed to this function. If set to False, a deep copy is
                made and altered instead (this is much slower).

        Returns:
            event (dict or object) with pedestals corrected data.

        Raises:
            BadDataError if the event is bad.
        """
        if not correct_in_place:
            event = copy.deepcopy(event)

        if self.pedestals is None:
            return event

        if not isinstance(event, dict):
            raise BadDataError(f"Not a valid event, got type: {type(event)}")

        if "data" not in event.keys():
            LOGGER.debug("Event is not parsed, didn't contain a data field.")
            return event

        corr_event = self._correct_pedestals(event)

        return corr_event

    def _correct_pedestals(self, event) -> dict:
        """Subtracts pedestals from a raw event.

        Corrects te pedestals using numpy for a significant speed boost.
        The function modifies the supplied event.
        This function must be run before any further post-processing.

        Args:
            event(dict): Target to subtract pedestals from.
        """
        if "data" in event.keys():
            evtdata = event["data"]
            window_labels = event["window_labels"]

            event["data"] = self._create_pedestals_corrected_data(
                evtdata, window_labels
            )
            event["pedestals_corrected"] = True
        else:
            LOGGER.debug("Event does not contain a data field.")
        return event

    def _create_pedestals_corrected_data(self, evtdata, window_labels):
        """Correct the data with the board.pedestals

        By fetching the window labels from the pedestals and applying the data to
        the event the pedestals are corrected. The pedestals contains corrections
        for all samples in all windows.

        Args:
            evtdata (np.array): With the data to be corrected.
            window_labels (list): List of window labels to pedestal correct.

        Returns:
            Pedestals corrected data in the same shape and format as before
        """
        samples = self.params["samples"]
        channels = self.params["channels"]

        pedcorrected = list()

        for chan in range(channels):
            winds = window_labels[chan]
            num_winds = len(winds)
            if num_winds == 0:
                pedcorrected.append(np.array([], dtype="float"))
                continue

            ped_locs = (samples * np.repeat(winds, samples)) + (
                np.ones((num_winds, samples), dtype=int)
                * np.arange(0, samples, dtype=int)
            ).flatten()
            ped_locs = ped_locs.astype(np.int64)

            p1 = self._pedestals[chan].take(ped_locs, mode="wrap")
            p2 = evtdata[chan]
            corrected = p2 - p1
            pedcorrected.append(corrected)

        pedcorrected = np.array(pedcorrected)

        return pedcorrected



import logging

import numpy as np

from naludaq.controllers import get_readout_controller
from naludaq.helpers.exceptions import OperationCanceledError
from naludaq.tools.pedestals.generators.default import PedestalsGenerator

from .default import PedestalsGenerator

LOGGER = logging.getLogger("naludaq.pedestals_generator_udc16")


class PedestalsGeneratorHiper(PedestalsGenerator):
    def __init__(
        self, board, num_captures: int, num_warmup_events: int, channels: list[int]
    ):
        self._board = board
        chan_per_chip = self._get_chan_per_chip()
        channels = channels or [
            x * chan_per_chip + y
            for x in self.board.params["installed_chips"]
            for y in range(chan_per_chip)
        ]
        super().__init__(board, num_captures, num_warmup_events, channels)
        trigger_limit = self.board.params.get("pedestals", {}).get("trigger_limit", 2.0)
        self._dc.trigger_interval_limit = trigger_limit

    def _get_chan_per_chip(self):
        # Convert channels into a channels per chip
        nchips = self.board.params.get("num_chips", 14)
        nchannels = self.board.params.get("channels", 56)
        return nchannels // nchips

    def _restore_backup_settings(self, backup_settings: dict):
        """Restore all backuped settings to the board.

        Returns:
            True if settings have been restored, False if no old settings were found.
        """
        if not backup_settings:
            return

        channels = backup_settings.get("readout_channels", None)
        if channels is not None:
            get_readout_controller(self.board).set_readout_channels(channels)

    def _validate_event(self, event: dict, expected_window_labels: list[int]) -> bool:
        """Returns true if the window labels matches the expected window labels.

        The validation matches the received window_labels with the expected
        window labels to make sure the block contains only data from that
        block. The firmware buffer sometimes contains more data from a previous block
        this makes sure it doesn't enter the next block buffer.

        Args:
            event (dict): event to validate
            expected_block (list): a list of expected window numbers

        Returns:
            True if validated, False if the events windows doesn't match expected.
        """
        for test_channel in self.channels:
            try:
                is_bad = False
                labels = np.array(event["window_labels"][test_channel])
                if not np.all(labels == expected_window_labels):
                    is_bad = True
                    LOGGER.warning(
                        "Channel %s.Expected: %s != returned: %s",
                        test_channel,
                        expected_window_labels,
                        labels,
                    )
                if len(event["data"][test_channel]) == 0:
                    LOGGER.warning(
                        "Missing data, channel %s, only has %s datapoints",
                        test_channel,
                        len(event["data"][test_channel]),
                    )
            except Exception as error_msg:
                LOGGER.error(f"Event validation failed due to: {error_msg}")
                return False
            if is_bad:
                LOGGER.warning("Event didn't pass validation")
                return False
        return True

    def generate_pedestals(self):
        """Generates pedestals and stores them in the board.pedestals.

        Pedestals stored on the board, they will be used on the next acquisition.

        If canceled, the board is set back to its previous state, and no pedestals
        are generated.
        """
        self._cancel = False
        old_pedestals = self.board.pedestals
        self._debug_raw_events = []

        self._update_progress(0, "Saving board state")
        backup = self._backup_settings()

        self.reset_pedestals()
        self._store_board_metadata()

        try:
            self._generate_pedestals_metadata_pre()
            blocks = self._capture_data_for_pedestals()
            self._generate_pedestals_metadata_post()
        except (OperationCanceledError, KeyboardInterrupt):
            self.board.pedestals = old_pedestals
        except:
            self.board.pedestals = old_pedestals
            raise
        else:  # else runs before finally if no error is raised
            self._update_progress(80, "Processing data")
            self._store_raw_data(blocks)
            self._store_averaged_data()
        finally:
            self._update_progress(90, "Restoring board state")
            self._restore_backup_settings(backup)



"""IO module

Collection of IO functions to support disk IO.
"""
import csv
import logging
import os

import numpy as np

LOGGER = logging.getLogger(__name__)


def get_csv_writer(params: dict):
    """Get the correct CSV writer for the board.

    Args:
        params (dict): board params.
        time_per_channel (bool): export time per channel rather than per row.
        include_timing (bool): include timing information in the CSV file.

    Returns:
        CSVWriter: CSV writer object.
    """
    try:
        model = params["model"]
    except Exception:
        model = None
    writer = CSVWriter
    if model in ["upac96", "udc16"]:
        writer = UpacCsvWriter

    return writer()


class CSVWriter:
    """
    CSV writer writes to CSV files.
    The CSV writer technically is two parts, a writer and a parser
    The parser is responsible for parsing the data into a format that the writer can write to disk.

    Attributes:
        time_per_channel (bool): export time per channel rather than per row.
        include_timing (bool): include timing information in the CSV file.
    """

    def export_acq_to_csv(
        self, filename: str, acqs: list, max_channels=None, time_per_channel=False
    ):
        """Exporting multiple acquisitions to a csv file.
        Args:
            filename (str): full filepath inlcuding directory and filename.
            acqs (list): list of acquisitions to export data from.
            max_channels (int): set the maximum channels to readout.
            time_per_channel (bool): export time per channel rather than per row.

        Output:
            csv file written to disk.
        """
        LOGGER.debug("Exporting %s acquisitions to csv", len(acqs))

        path, _ = os.path.split(filename)
        if not os.path.isdir(path):
            raise NotADirectoryError(f"Not a valid directory: {path}")
        if not isinstance(acqs[0], dict):
            events = self._compress_acquisitions(acqs)
        else:
            events = acqs
        self._export_acq_to_csv_per_event(
            filename, events, max_channels, time_per_channel
        )

    def _export_acq_to_csv_per_event(
        self, filename: str, events: list, max_channels=None, time_per_channel=False
    ):
        """Iterate over a list of events using a np.array buffer to write to a csv file on disk. Use if
        events are the same in size.

        Generate the event matrix by realigning the columns based on the window label.
        There is a master label list and all the other will be moved accordingly
        This means that the first label in the master label is used as a relative 0 x-axis
        All other data is moved accordngly

        Args:
            filename (str): full filepath inlcuding directory and filename.
            events (list): list of events to export data from.
            max_channels (int): set the maximum channels to readout.
            time_per_channel (bool): export time per channel rather than per row.

        Output:
            csv file written to disk.
        """
        num_events = len(events)
        num_channels = self._calculate_num_channels(events)
        include_timing = False
        decimals_to_round = 1
        # Override time_per_channel if window labels are not uniform in each event.
        if not self._check_window_label_uniformity(events):
            time_per_channel = True
        if events[0].get(
            "timing", False
        ):  # If data contains timing field, it must be included per channel
            time_per_channel = True
            include_timing = True

        csv_header = self._create_csv_header(
            num_channels, time_per_channel, include_timing
        )
        with open(filename, "w", newline="") as evt_file:
            header = ",".join(csv_header) + "\n"
            evt_file.write(header)

        num_columns = len(csv_header)  # Columns is common for all events

        for i, evt in enumerate(events):
            if "data" not in evt:
                LOGGER.debug("Event %s has no data, skipping", i)
                continue
            samples = self._calculate_sp_per_chan(evt)
            labels = self._calculate_window_labels(
                evt
            )  # Will produce errors on HDSoC no matter what
            rows = len(labels) * samples

            np_array = np.full(shape=(rows, num_columns), fill_value="", dtype=object)

            # write evt & acq column
            acq_col = 0
            evt_num_col = 1
            np_array[:, acq_col] = evt.get("acq_num", 0)
            np_array[:, evt_num_col] = evt.get("pkg_num", i)

            if not time_per_channel:
                window_label_col_idx = 2
                window_label_column = self._generate_label_column(labels, samples)
                np_array[:, window_label_col_idx] = window_label_column
            col_idx = 0
            for idx, data in enumerate(evt.get("data", list())):
                if (
                    len(data) == 0
                ):  # Channel isn't enabled if it has no corresponding window labels
                    continue
                xaxis = evt["time"][idx]
                win_labels = evt["window_labels"][idx]
                data_rows = self._get_start_end(
                    data, win_labels, labels, samples
                )  # Generates issues with HDSoC

                (
                    window_label_col_idx,
                    data_col_idx,
                    time_col_idx,
                    timing_label_col_idx,
                ) = self._generate_column_indexes(
                    col_idx, time_per_channel, include_timing
                )

                if time_per_channel:
                    window_label_column = self._generate_label_column(
                        win_labels, samples
                    )
                    np_array[data_rows, window_label_col_idx] = window_label_column

                if include_timing:
                    timings = evt["timing"][idx]
                    timings_column = self._generate_label_column(timings, samples)
                    np_array[data_rows, timing_label_col_idx] = timings_column

                if not isinstance(data[0], int):
                    data = np.round(
                        np.array(data).astype(float), decimals=decimals_to_round
                    )
                col_idx += 1

                np_array[data_rows, data_col_idx] = data
                np_array[data_rows, time_col_idx] = xaxis

            with open(filename, "a", newline="", buffering=10_000_000) as evt_file:
                np.savetxt(evt_file, np_array, fmt="%s", delimiter=",")

        LOGGER.info("Finished writing %d events to CSV", num_events)

    def _generate_label_column(self, labels: list[int], samples: int) -> np.array:
        """Generate a column of window labels for a single event.

        Window numbers start with the earliest window number then continues until all numbers have been listed.
        There can be holes in the window number but data will not "Wrap around", window labels will continue.

        Args:
            labels (list): list of window labels.
            samples (int): number of samples per window.

        Returns:
            np.array: array of window labels.
        """
        num_window_labels = len(labels)
        window_label_column = np.reshape(
            np.repeat(labels, samples), [samples, num_window_labels]
        ).flatten()
        return window_label_column

    def _generate_column_indexes(
        self, idx, time_per_channel: bool = True, include_timing: bool = False
    ) -> "tuple[int, int, int, int|None]":
        """Generate the column indexes for the np.array based on the time_per_channel flag.

        The two first columns are always the same, the rest is dependent on the time_per_channel flag.

        If time_per_channel is True, every three columns, are:
        window_label_X, data_X, time_X where X is the channel number.

        If time_per_channel is False, the third column is the time column and the rest are data columns.

        Args:
            time_per_channel (bool): export time per channel rather than per row.
            idx (int): index of the channel.
            include_timing (bool): include timing information in the CSV.

        Returns:
            tuple: window_label_col_idx, data_col_idx, time_col_idx
        """
        if time_per_channel:
            window_label_col_idx = 2 + ((3 + include_timing) * idx)
            data_col_idx = window_label_col_idx + 1
            time_col_idx = window_label_col_idx + 2
            timing_col_idx = window_label_col_idx + 3 if include_timing else None
        else:
            window_label_col_idx = 2
            data_col_idx = 4 + idx
            time_col_idx = 3
            timing_col_idx = None
        return window_label_col_idx, data_col_idx, time_col_idx, timing_col_idx

    def _get_start_end(self, data, win_labels, labels, samples) -> slice:
        """Get the start and end of the data in the np.array.

        Args:
            data (list): list of data.
            win_labels (list): list of window labels.
            labels (list): list of labels to export.
            samples (int): number of samples per window.

        Returns:
            slice: start and end of the data in the np.array.

        Raises:
            IndexError: if the label doesn't exist in the merged labels.
        """
        # create a data array with the correct number of rows and np.nan
        label_pos = np.where(win_labels[0] == labels)
        try:
            label_pos = (label_pos[0].astype(int) * samples)[0]
        except IndexError:
            LOGGER.debug("Label doesn't exist in the merged labels.")
            label_pos = 0

        start = label_pos
        end = start + len(data)
        return slice(start, end)

    @staticmethod
    def _compress_acquisitions(acquisitions):
        """Converts multiple acquisitions into a single
        acquistion.

        Args:
            acquisitions (list): A list of acquisition

        Returns:
            acquisition (list): Combined acquisition of all events
        """
        events = list()
        for acq in acquisitions:
            events.extend(acq)

        return events

    @staticmethod
    def _calculate_max_channels(events):
        """Calculates the maximum channels that exists of
        all events.

        Args:
            events (list): List of events, or acq

        Returns:
            max_chan (int): Max channels from all events
        """
        return max([len(x.get("data", list())) for x in events])

    def export_pedestals_csv(self, board, filename: str, pedestals: dict):
        """Export all data from the pedestals to CSV.

        Please note this will only export the data, not the information regarding the pedestals.
        This function also assumes filename is in the correct format.
        It will override any existing file.

        Args:
            board (Board): A board object with the pedestals data attached to it.
            filename (path): Filename for the csv file.
        """
        path, _ = os.path.split(filename)
        if not os.path.isdir(path):
            raise NotADirectoryError(f"Not a valid directory: {path}")

        max_channels = self._calculate_max_channels([pedestals])
        _csv_header = self._create_csv_header(max_channels, False)

        with open(filename, mode="w", newline="") as evt_file:
            writer = csv.writer(evt_file, delimiter=",")
            writer.writerow(_csv_header)
            peds_shape = pedestals["data"].shape
            sample_count = peds_shape[2]
            # sample_count = board.params["samples"]
            window_count = peds_shape[1]
            channels_count = peds_shape[0]
            for window in range(window_count):
                for sample in range(sample_count):
                    row = []
                    row.append(-1)  # acq number
                    row.append(-1)  # evt number
                    row.append(window)
                    row.append(window * sample_count + sample)  # timing

                    for channel in range(channels_count):
                        row.append(round(pedestals["data"][channel][window][sample], 2))
                    writer.writerow(row)

    @staticmethod
    def _create_csv_header(
        max_channels: "int|list",
        time_per_channel: bool = True,
        include_timing: bool = False,
    ) -> list[str]:
        """Creates the csv file header depending on number of channels.

        The determine the amount of channels the csv file need to handle.

        Args:
            max_channels(int|list): list of channels or amount of channels in the header.
            time_per_channel(bool): Add time and window columns per channel
            pedestals(bool): If true, add column for pedestals

        Returns:
            list of header labels
        """
        if isinstance(max_channels, int):
            channels = range(max_channels)
        elif isinstance(max_channels, list):
            channels = max_channels
        else:
            raise TypeError("max_channels must be an int or list")

        csv_header = ["acq_num", "evt_num"]
        if time_per_channel is False:
            csv_header.append("windnum")
            csv_header.append("time")
        for index in channels:
            if time_per_channel is True:
                csv_header.append("windnum_ch" + str(index))
            csv_header.append("data_ch" + str(index))
            if time_per_channel is True:
                csv_header.append("time_ch" + str(index))
            if include_timing is True:
                csv_header.append("timing_ch" + str(index))

        return csv_header

    def _calculate_num_channels(self, events) -> int:
        """Calculates the unique of channels in a list of events.

        Args:
            events (list): List of events

        Returns:
            int: Number of channels
        """

        if isinstance(events, dict):
            events = [events]
        channels = set()
        for event in events:
            data = event.get("data", None)
            if data is None:
                continue
            evt_chans = [i for i, d in enumerate(data) if len(d) > 0]
            channels.update(evt_chans)
        return list(channels)

    def _calculate_sp_per_chan(self, event):
        """Calculates the number of samples per channel in an event.

        It used the amount of window labels and the number of samples to calculate
        the number of samples per channel.
        """
        labels = [(i, x) for i, x in enumerate(event["window_labels"]) if len(x) > 0]
        ch = labels[0][0]  # Guaranteed data
        num_samps = len(event["data"][ch]) // len(labels[0][1])
        return num_samps

    def _calculate_window_labels(self, event: dict) -> list[int]:
        """Merges the window labels in order.

        Args:
            event: standard event format
            include_timing: include timing to determine start window

        Returns:
            A list of merged labels in order of appearance.
        """
        output = []
        nextround = [x for x in event["window_labels"] if len(x) > 0]
        while nextround:
            longest_remaining = max([len(x) for x in nextround])
            if longest_remaining == 0:
                break
            topop = CSVWriter()._calculate_first_window_idx(nextround)
            nxt_lbl = nextround[topop][0]
            if output:  # Guard against empty list
                if output[-1] != nxt_lbl:  # Guard against duplicates
                    output.append(nxt_lbl)
            else:
                output.append(nxt_lbl)
            available_labels = [y for y in nextround if len(y) > 0]
            nextround = [x[1:] if x[0] == nxt_lbl else x for x in available_labels]

        return output

    def _calculate_first_window_idx(self, alabels: list[list[int]]) -> int:
        """Calculates the index of the first window in the list of labels.

        Args:
            alabels: list of labels

        Returns:
            int: index of the first window
        """
        rollvers = [
            (p, i[0])
            for p, i in [(p, x) for (p, x) in enumerate(alabels) if len(x) > 0]
            if i[0] > i[-1]
        ]
        no_rollover = [
            (p, i)
            for p, i in [(p, x) for (p, x) in enumerate(alabels) if len(x) > 0]
            if i[0] < i[-1]
        ]

        first = []
        if rollvers:  # channel with rollovers
            # check if first is in the remaining alabels
            # if it is, then it can't be the first window
            output = []
            for f in rollvers:
                for n in no_rollover:
                    if f[1] in n[1]:
                        break
                    output.append(f)

            first = output

        if not first:
            first = [(pos, x[0]) for pos, x in enumerate(alabels) if len(x) > 0]

        topop = min(first, key=lambda x: x[1])  # smallest of the visible
        return topop[0]

    def _calculate_first_chan(self, event):
        """USes the timing labels to determine the first channel sampled.

        NNTE: lowest timing = first sampled data is an UNCONFIRMED assumption!
        """
        timings = [(i, x[0]) for i, x in enumerate(event["timing"]) if len(x) > 0]
        mintim = min(timings, key=lambda x: x[1])
        minidx = mintim[0]
        return event["window_labels"][minidx][0]

    def _check_window_label_uniformity(self, events: list[dict]) -> bool:
        """Will return true if all channels in every events in a list of events have the same window labels."""
        for event in events:
            labels = event.get("window_labels", None)
            if labels is None:
                continue
            labels = [x for x in labels if len(x) > 0]
            chans = set()
            for ch in labels:
                chans.update(ch)
            for ch in labels:
                if len(ch) != len(chans):
                    return False

        return True

    def export_thresholdscan_to_csv(
        self,
        filename: str,
        trigger_values,
        threshold_results: np.array,
        channels: "list|None" = None,
    ):
        """Write the results of a threshold scan to a csv file.
        The file will be formatted such that rows correspond to threshold values and columns correspond to channels

        Expected file output:
            Trigger value,ch0,ch1,ch2,...
            5,0,0,0,...
            10,2000,2000,0,...
            ...

        Args:
            filename (str): full filepath of of csv file to write to
            trigger_values (Iterable[Int]): contains the values which the threshold value were triggered on
            threshold_outputs (np.array): An array of shape len(trigger_values) x num_channels
            channels (list[int] or None): List of channels to include, or None to include all channels

        Output:
            thresholdscan outputs to csv file
        """
        try:
            threshold_results.shape
        except AttributeError:
            raise TypeError(
                f"threshold_results should be a numpy array, got {type(threshold_results)}"
            )
        if channels is not None:
            channels = sorted(channels)
            threshold_results = threshold_results[:, tuple(channels)]
        else:
            channels = range(threshold_results.shape[1])

        if len(trigger_values) != threshold_results.shape[0]:
            raise ValueError(
                "The length of trigger values and thresholdscan values should match ({} vs {})".format(
                    len(trigger_values), threshold_results.shape[0]
                )
            )

        header = ["Trigger value"] + ["ch{}".format(x) for x in channels]
        with open(filename, "w", newline="") as dst:
            writer = csv.writer(dst)
            writer.writerow(header)
            for i, trig_val in enumerate(trigger_values):
                writer.writerow([trig_val, *threshold_results[i]])


class UpacCsvWriter(CSVWriter):
    def _export_acq_to_csv_per_event(
        self,
        filename: str,
        events: list,
        max_channels=None,
        time_per_channel: bool = True,
    ):
        """Iterate over a list of events using a np.array buffer to write to a csv file on disk. Use if
        events are the same in size.

        Generate the event matrix by realigning the columns based on the window label.
        There is a master label list and all the other will be moved accordingly
        This means that the first label in the master label is used as a relative 0 x-axis
        All other data is moved accordngly

        Args:
            filename (str): full filepath inlcuding directory and filename.
            events (list): list of events to export data from.
            max_channels (int): set the maximum channels to readout.
            time_per_channel (bool): export time per channel rather than per row.
                The value is ignored since the UDC data format requires time per window.

        Output:
            csv file written to disk.
        """
        time_per_channel = True
        num_events = len(events)
        num_channels = self._calculate_num_channels(events)

        decimals_to_round = 1

        csv_header = self._create_csv_header(
            num_channels,
            time_per_channel=time_per_channel,
            include_timing=False,
        )
        with open(filename, "w", newline="") as evt_file:
            header = ",".join(csv_header) + "\n"
            evt_file.write(header)

        num_columns = len(csv_header)  # Columns is common for all events

        for i, evt in enumerate(events):
            if "data" not in evt:
                LOGGER.debug("Event %s has no data, skipping", i)
                continue
            samples = self._calculate_sp_per_chan(evt)
            labels = max([len(x) for x in evt.get("window_labels", []) if len(x) > 0])
            rows = labels * samples

            np_array = np.full(shape=(rows, num_columns), fill_value="", dtype=object)

            # write evt & acq column
            acq_col = 0
            evt_num_col = 1
            np_array[:, acq_col] = evt.get("acq_num", 0)
            np_array[:, evt_num_col] = evt.get("pkg_num", i)

            if not time_per_channel:
                window_label_col_idx = 2
                window_label_column = min(
                    [x[0] for x in evt.get("window_labels", []) if len(x) > 0]
                )
                np_array[:, window_label_col_idx] = window_label_column
            col_idx = 0
            for idx, data in enumerate(evt.get("data", list())):
                if (
                    len(data) == 0
                ):  # Channel isn't enabled if it has no corresponding window labels
                    continue
                xaxis = evt["time"][idx]
                win_labels = evt["window_labels"][idx]
                data_rows = slice(0, len(data))
                (
                    window_label_col_idx,
                    data_col_idx,
                    time_col_idx,
                    timing_label_col_idx,
                ) = self._generate_column_indexes(
                    col_idx, time_per_channel, include_timing=False
                )

                if time_per_channel:
                    window_label_column = self._generate_label_column(
                        win_labels, samples
                    )
                    np_array[data_rows, window_label_col_idx] = window_label_column

                if not isinstance(data[0], int):
                    data = np.round(
                        np.array(data).astype(float), decimals=decimals_to_round
                    )
                col_idx += 1

                np_array[data_rows, data_col_idx] = data
                np_array[data_rows, time_col_idx] = xaxis

            with open(filename, "a", newline="", buffering=500_000) as evt_file:
                np.savetxt(evt_file, np_array, fmt="%s", delimiter=",")

        LOGGER.info("Finished writing %d events to CSV", num_events)



import logging
from typing import List, Tuple

from naludaq.communication import AnalogRegisters
from naludaq.helpers.exceptions import DACError

from .base import BaseTIAController

LOGGER = logging.getLogger("naludaq.tia_controller_hdsoc")


class HdsocTIAController(BaseTIAController):
    def __init__(self, board, precision: float = 3.0):
        """Controller for the TIA on the HDSoC board.

        Sets the TIA DAC using dynamic precision. Upper/lower reference voltages
        are calculated to maximize the precision of the values set. This means that
        setting the TIA DACs to a wide range of values results in a lower precision,
        hence the minimum "precision" argument. Therefore the best results are obtained
        when the DAC values for all channels are relatively close, or ideally,
        the same value.

        Args:
            board (Board): the board object.
            precision (float): the minimum level of precision allowed. Roughly the
                maximum amount the given DAC values and the actual DAC values may
                differ by.
        """
        super().__init__(board)
        self.precision = precision
        self._valid_sides = ["left", "right"]

        params = self.board.params.get("tia_dac", {})
        self.lower_limit = params.get("min_vref", 0)
        self.upper_limit = params.get("max_vref", 2500)
        self._interval = self.upper_limit - self.lower_limit
        self._steps = 2 ** params.get("val_bits", 8) - 1
        self._num_subranges = 2 ** params.get("ref_bits", 4) - 1
        self._ref_table = {
            i: int((self.upper_limit - self.lower_limit + 1) / self._num_subranges * i)
            for i in range(self._num_subranges + 1)
        }

    @property
    def lower_limit(self) -> int:
        """Get the lower value limit"""
        return self._lower_limit

    @lower_limit.setter
    def lower_limit(self, value: int):
        if not isinstance(value, int):
            raise TypeError("lower_limit must be an int")
        if (
            not self.board.params.get("min_vref", 0)
            <= value
            <= self.board.params.get("max_vref", 2500)
        ):
            raise ValueError("Lower limit must be min_vref < value < max_vref")
        self._lower_limit = value

    @property
    def upper_limit(self) -> int:
        """Get the upper value limit"""
        return self._upper_limit

    @upper_limit.setter
    def upper_limit(self, value: int):
        if not isinstance(value, int):
            raise TypeError("lower_limit must be an int")
        if (
            not self.board.params.get("min_vref", 0)
            <= value
            <= self.board.params.get("max_vref", 2500)
        ):
            raise ValueError("Lower limit must be min_vref < value < max_vref")
        self._upper_limit = value

    @property
    def precision(self) -> float:
        """Get/set the minimum precision.

        Raises:
            TypeError if not set to a numeric value.
            ValueError if the value is not positive.
        """
        return self._precision

    @precision.setter
    def precision(self, value: float):
        if not isinstance(value, (int, float)):
            raise TypeError("Precision must be numeric")
        if value <= 0:
            raise ValueError("Precision must be a positive number")
        self._precision = value

    def set_enabled_sides(self, sides: "str | list[str]"):
        """Sets which TIAs are enabled.

        **Important:** the external DAC must be disabled in order
        to enable either of the TIAs.

        Args:
            sides (str | List[str]): a list of sides to enable, or a str
                for a single side. Sides must be "left" or "right" (case sensitive).

        Raises:
            TypeError if an invalid type is given.
            ValueError if the sides are not "left" or "right"
            DACError if the external DAC is not disabled.
        """
        sides = list(set(sides)) if isinstance(sides, list) else [sides]
        if not all(isinstance(x, str) for x in sides):
            raise TypeError('Sides must be strings "left" or "right"')
        if not all(x in ["left", "right"] for x in sides):
            raise ValueError('Side must be "left" or "right"')
        if len(sides) > 0 and self._is_ext_dac_enabled():
            raise DACError("External DAC must be disabled to use the TIA")

        for side in self._valid_sides:
            enabled = side in sides
            self._write_analog_register(f"ntia_{side}", not enabled)

    def set_dac_values(self, values: list[int]):
        """Sets the TIA DAC values for all channels.

        Writes the DAC values with dynamic precision, meaning that
        what you get isn't necessarily what you requested, up to
        a certain tolerance level. This function maximizes the
        level of precision you get, but the precision quickly drops
        as the dac values become more spread out. See the class
        docstring for more information.

        Args:
            values (List[int]): the TIA DAC values. Bounds: lower limit (in mV) to upper limit (in mV)

        Raises:
            TypeError if `values` is not a list of ints.
            ValueError if one or more values is out of bounds.
        """
        self._validate_values(values)

        min_dac, max_dac = min(values), max(values)
        low_ref, high_ref = self._get_references(min_dac, max_dac)
        low_vref = self._ref_table[low_ref]
        high_vref = self._ref_table[high_ref]
        high_vref = min(high_vref, self._upper_limit)

        actual_precision = self._calculate_precision(low_vref, high_vref)
        if self.precision < actual_precision:
            raise ValueError(
                f"Cannot set TIA DAC values: {min_dac} and {max_dac} are too far apart. "
                f"The precision ({actual_precision} mV) is lower than the minimum allowed "
                f"({self.precision} mV)"
            )

        self._set_references(low_ref, high_ref)
        self._set_tia_dac_values(values, low_vref, high_vref)

    def _get_references(self, lower_limit: int, upper_limit: int) -> Tuple[int, int]:
        """Return the 4-bit reference value for upper and lower trigger reference voltages.

        Args:
            lower_limit(int): Lowest trigger value
            upper_limit(int): Highest trigger value

        Returns:
            A tuple of (low_vref, high_vref)

        Raises:
            TypeError if the lower/upper limits are not an int
            ValueError if one or more values is out of bounds, or the
                minimum exceeds the maximum.
        """
        self._validate_bounds(
            lower_limit, upper_limit, self._lower_limit, self._upper_limit
        )
        low_ref = int(divmod(lower_limit, self._interval / self._num_subranges)[0])
        high_ref = low_ref + 1  # Upper reference above the value

        # Sanity checks. Refs can't be equal, so we need to slide one of them up/down
        # Only equal when hitting the top of the range.
        high_ref = min(high_ref, self._num_subranges)
        if low_ref == high_ref:
            low_ref -= 1
        return low_ref, high_ref

    def _set_references(self, min_vref: int, max_vref: int):
        """Set the Vdd and Vss references for both east and west side.

        Args:
            min_vref(int): the lower value
            max_vref(int): the upper value

        Raises:
            TypeError if the min/max refs are not ints
            ValueError if one or more arguments is out of bounds, or if
                the minimum exceeds the maximum.
        """
        self._validate_bounds(min_vref, max_vref, 0, self._num_subranges)
        if min_vref == max_vref:
            raise ValueError("Minimum and maximum references cannot be the same")
        self._write_analog_register("sub_ref_neg_sb_left_tia", min_vref)
        self._write_analog_register("sub_ref_neg_sb_right_tia", min_vref)
        self._write_analog_register("sub_ref_pos_sb_left_tia", max_vref)
        self._write_analog_register("sub_ref_pos_sb_right_tia", max_vref)

    def _set_tia_dac_values(self, values: List[int], low_vref: int, high_vref: int):
        """Writes the corrected DAC values to the corresponding registers.

        Args:
            values (List[int]): the list of values for each channel.
            low_vref (int): the lower vref.
            high_vref (int): the higher vref.

        Raises:
            TypeError if the values are not a List[int] or the low/high refs
                are not ints.
            ValueError if any values are out of bounds, or if the low reference
                exceeds the high reference.
        """
        self._validate_values(values)
        self._validate_bounds(low_vref, high_vref, self._lower_limit, self._upper_limit)
        step_size = (high_vref - low_vref) / self._steps
        for channel, value in enumerate(values):
            corrected_value = round(max(value - low_vref, 0) / step_size)
            self._write_analog_register(f"tia_dac_{channel}", corrected_value)

    def _is_ext_dac_enabled(self) -> bool:
        """Checks if any of the external DACs are enabled for any channels."""
        return any(x != "hi-z" for x in self.board.dac_values.values())

    def _calculate_precision(self, low_vref: float, high_vref: float):
        """Calculates the highest level of precision we can get
        with the given reference voltages.

        Args:
            low_vref (float): the lower vref
            high_vref (float): the higher vref

        Returns:
            The precision
        """
        return (high_vref - low_vref) / self._steps

    def _validate_values(self, values):
        """Validates a value expected to be a valid list of ints.

        Args:
            values: the values to check

        Raises:
            TypeError if the argument is not a List[int]
            ValueError if the list has the wrong length or
                any values are out of bounds.
        """
        if not isinstance(values, list):
            raise TypeError("Values must be a List[int]")
        if not len(values) == self._board.channels:
            raise ValueError(
                f"Values must be an array of length {self._board.channels}"
            )
        if not all(isinstance(x, int) for x in values):
            raise TypeError("Values must be a List[int]")
        if not all(self._lower_limit <= x <= self._upper_limit for x in values):
            raise ValueError("One or more values is out of bounds.")

    def _validate_bounds(self, given_low, given_high, min, max):
        """Validates the low and high bounds.

        Args:
            given_low (int): the given lower bound
            given_high (int): the given upper bound
            min (int): minimum allowed value
            max (int): maximum allowed value

        Raises:
            TypeError if either argument is not an int
            ValueError if either vref is out of bounds,
                or if the low vref exceeds the high vref.
        """
        if not isinstance(given_low, int):
            raise TypeError("low_vref must be an int")
        elif not isinstance(given_high, int):
            raise TypeError("high_vref must be an int")
        elif not min <= given_low <= max:
            raise ValueError("low_vref is out of bounds")
        elif not min <= given_high <= max:
            raise ValueError("high_vref is out of bounds")

    def _write_analog_register(self, register: str, value: int):
        """wrapper for the Analog register coms module.

        Args:
            register (str): name of the register to update.
            value (int): The register value to set.

        Raises:
            ValueError: if the value cannot be written
            TypeError: if the register/value are of incorrect types.
        """
        AnalogRegisters(self.board).write(register, value)



from .default import DefaultAutoTrigger


def get_autotrigger(board, interval: float = 1) -> DefaultAutoTrigger:
    """Gets the autotrigger module for the given board.

    Args:
        interval (float): the interval to wait between sending triggers.
    """
    # this weird code is intended for forward-compatibility
    return {}.get(board.model, DefaultAutoTrigger)(board, interval)



"""Board controller specific for HDSoC series of boards
"""
from logging import getLogger
from typing import List

from naludaq.communication import AnalogRegisters, ControlRegisters
from naludaq.helpers.register_cache import RegisterCache
from naludaq.helpers.semiton import SemitonABC

from .default import BoardController

LOGGER = getLogger("naludaq.board_controller_hdsoc")


class HDSoCBoardController(BoardController, SemitonABC):
    """Board controller for HDSoCv1.

    Only one instance of this class per board object is allowed.
    """

    def __init__(self, board):
        super().__init__(board)
        self._channels_per_chip = board.params.get("channels_per_chip", 16)
        self._num_chips = board.params.get("num_chips", 2)
        self._chip_states = {chip: True for chip in range(self._num_chips)}
        self._waittime = 0.01
        self._init_power_state_caches()

    def _init_power_state_caches(self):
        """Initializes the analog register cache used to remember
        the values of registers after they are ovewritten when
        turning off the chip.
        """
        self._disable_register_states = {
            "channel_bias_bias_{}": 0,
            "scvbias_{}": 0xFFF,
            "scvstab_{}": 0,
            "channel_dac_bias_{}": 0,
            "channel_dac_bias_bias_{}": 0,
            "cmpbias_{}": 0,
            "pubias_{}": 0xFFF,
            "cmpbias2_{}": 0,
            "isel_{}": 0,
            "ref_output_bias_{}": 0,
            "ref_output_bias_bias_{}": 0,
        }

        self._power_state_caches: List[RegisterCache] = []
        for side in range(self._num_chips):
            side_name = {0: "left", 1: "right"}[side]
            cache = RegisterCache(self._board)
            for name in self._disable_register_states:
                name = name.format(side_name)
                cache.add(name, AnalogRegisters)
            self._power_state_caches.append(cache)

    def stop_readout(self):
        """Toggles the "stopacq" signal on the readout module.

        It's equivalent of asking it nicely to stop reading.
        """
        super().stop_readout()
        self.clear_buffer()

    def digital_reset(self):
        """Toggles the "reset" port on the readout module.

        Forcibly returns the chip to default state.
        """
        self._write_control_register("idig_rst", True)
        self._write_control_register("idig_rst", False)

    def clear_buffer(self):
        """Clears the UART buffer on both CPU and FPGA side."""
        self._clear_fpga_buffer()
        super().clear_buffer()

    def _clear_fpga_buffer(self):
        """Resets the FPGA FIFO"""
        ControlRegisters(self.board).write("wave_fifo_rst", True)
        ControlRegisters(self.board).write("wave_fifo_rst", False)
        self.digital_reset()

    #############################################################################
    # Random Stuff, may or maynot work... Below functions need to be validated.
    #############################################################################

    def set_enabled_chips(self, chips: List[int]):
        """Sets which chips are enabled.

        Args:
            chips (List[int]): the list of chips. Chip numbers can be
                repeated; it makes no difference in the register writes.
        """
        self._validate_chip_numbers_or_raise(chips)
        for chip in range(self._num_chips):
            self.set_chip_enabled(chip, chip in chips)

    def set_chip_enabled(self, chip_num: int, enabled: bool):
        """Turns on/off a lot of biases on one side of a specific chip.

        Args:
            chip_num (int): the chip number, 0 or 1.
            enabled (bool): whether the channel should be enabled or disabled.

        Raises:
            TypeError if the `chip_num` is not an int, or `enabled` not a bool.
            ValueError if the chip number is out of bounds.
        """
        self._validate_chip_numbers_or_raise([chip_num])
        if not isinstance(enabled, bool):
            raise TypeError('"enabled" must be a bool')

        if enabled:
            self._enable_chip(chip_num)
        else:
            self._disable_chip(chip_num)

    def _enable_chip(self, chip_num: int):
        """Turns on a lot of biases on one side of the chip.

        Args:
            chip_num (int): 0 or 1, which side to turn on biases

        Raises:
            TypeError if the argument is not an int.
            ValueError if the chip number is out of bounds.
        """
        self._validate_chip_numbers_or_raise([chip_num])
        self._power_state_caches[chip_num].restore_all()
        self._chip_states[chip_num] = True

    def _disable_chip(self, chip_num: int):
        """Turns off a lot of biases on one side of the chip to reduce power draw.

        Args:
            chip_num (int): 0 or 1, which side to turn off biases

        Raises:
            TypeError if the argument is not an int.
            ValueError if the chip number is out of bounds.
        """
        self._validate_chip_numbers_or_raise([chip_num])

        # Disabling the chip overwrites old values, need to cache to be able to restore
        if self._chip_states[chip_num]:
            self._power_state_caches[chip_num].update_all()
        self._write_analog_registers(self._disable_register_states, chip=chip_num)
        self._chip_states[chip_num] = False

    def _validate_chip_numbers_or_raise(self, chips: List[int]):
        """Validates a list of chip numbers

        Args:
            chips (List[int]): the list of chips

        Raises:
            TypeError if the argument is not a list of ints.
            ValueError if any chip number is out of bounds.
        """
        if not isinstance(chips, list):
            raise TypeError("Chip numbers must be a list")
        if not all(isinstance(x, int) for x in chips):
            raise TypeError("Chip numbers must all be ints")
        if len(chips) != 0:
            if min(chips) < 0:
                raise ValueError(f"Invalid chip number: {min(chips)}")
            if max(chips) >= self._num_chips:
                raise ValueError(f"Invalid chip number: {max(chips)}")

    def _write_analog_registers(self, registers: dict, chip=None):
        """Write several analog registers.

        Args:
            registers (dict): a dict of `reg_name: reg_value`.
            chip (int): the chip number (0, 1). If not `None`, a suffix
                '_left' or '_right' is appended to each register name.
            preserve (bool): whether to preserve the original value in the
                software registers after writing.
        """
        ar = AnalogRegisters(self.board)
        chip_suffix = {
            0: "left",
            1: "right",
        }
        for register, value in registers.items():
            reg_name = register.format(chip_suffix.get(chip, ""))
            ar.write(reg_name, value)



import abc
import typing
from collections import deque

from naludaq.backend import AcquisitionManager
from naludaq.daq import get_daq
from naludaq.helpers.exceptions import BadDataError
from naludaq.tools.waiter import EventWaiter


class DaqInterface(abc.ABC):
    def __init__(self, board: str):
        """Interface combining the old and new DAQ."""
        self._board = board

    @abc.abstractmethod
    def start_capture(self):
        """Start the data capture.

        Must be called before using stream()
        """

    @abc.abstractmethod
    def stop_capture(self):
        """Stop the data capture"""

    @abc.abstractmethod
    def stream(self, timeout: float) -> typing.Iterable[dict]:
        """Generator for valid events.

        This generator will continually yield valid events until
        the timeout is reached or the first invalid event is captured.

        The start_capture() method must be called before using this method.

        Args:
            timeout (float): timeout in seconds for each event.
                The timeout is reset for each event.

        Raises:
            TimeoutError: if no events are captured within the timeout
            BadDataError: if an invalid event is captured
        """


def get_daq_interface(board: str) -> DaqInterface:
    """Internal method for getting the correct DAQ interface.

    The DAQ interface allows an abstraction around the two DAQs which allows
    the user to be DAQ-agnostic until the old DAQ is sunset.
    """
    if board.using_new_backend:
        return NewDaqInterface(board)
    return OldDaqInterface(board)


class OldDaqInterface(DaqInterface):
    def __init__(self, board: str):
        super().__init__(board)
        self._output_buffer = deque()

        self._daq = get_daq(board, parsed=True)
        self._daq.output_buffer = self._output_buffer

    def start_capture(self):
        self._daq.start_capture()

    def stop_capture(self):
        self._daq.stop_capture()

    def stream(self, timeout: float) -> typing.Iterable[dict]:
        while True:
            # It's possible more than one event is in the buffer,
            # in which case we don't need to wait.
            if self._output_buffer:
                event = self._output_buffer.popleft()
                if "data" not in event:
                    raise BadDataError("Invalid event captured")
                yield event
                continue
            waiter = EventWaiter(self._output_buffer, 1, 0.01, timeout)
            try:
                waiter.start(blocking=True)
            except TimeoutError:
                raise


class NewDaqInterface(DaqInterface):
    def __init__(self, board: str):
        super().__init__(board)
        self._acq = None

    def __del__(self):
        try:
            self.stop_capture()
        except:
            pass

    def start_capture(self):
        self._acq = AcquisitionManager(self._board).create_temporary()
        self._acq.set_output()

    def stop_capture(self):
        try:
            self._acq.delete()
        except:
            # most likely means it was already deleted
            pass
        self._acq = None
        AcquisitionManager(self._board).current_acquisition = None

    def stream(self, timeout: float) -> typing.Iterable[dict]:
        start = 0
        skip_bad = False
        return self._acq.stream_parsed(timeout, skip_bad=skip_bad, start=start)



import logging

from naludaq.controllers.external_dac.dac7578 import DACControllerDAC7578

logger = logging.getLogger("naludaq.ext_dac.aardvarcv3")


class DacControllerAardvarcv3(DACControllerDAC7578):
    def __init__(self, board, send_delay=0.001) -> None:
        """External DAC controller for the AARDVARCv3."""
        super().__init__(board, send_delay)
        self._cal_address = self.board.params["ext_dac"]["cal"]["address"]
        self._cal_channel = self.board.params["ext_dac"]["cal"]["channel"]

    @property
    def cal_dac(self) -> int:
        """Get the calibration channel DAC value."""
        return self.board.params.get("ext_dac", {}).get("cal", {}).get("value", 0)

    @cal_dac.setter
    def cal_dac(self, value: int):
        """Set the calibration channel DAC value."""
        self.set_cal_dac(value, set_mv=False)

    def set_cal_dac(self, value: int, set_mv: bool = False):
        self._validate_value(value, set_mv)
        if set_mv:
            value = self._convert_mv2cnt(value)

        logger.debug("Setting calibration channel to %d counts", value)
        words = [
            f"{self._write_cmd:04b}{self._cal_channel:04b}",
            f"{(value >> 4 & 0xFF):08b}",
            f"{(value << 4 & 0xFF):08b}",
        ]
        addr = self._cal_address
        self._send_command(addr, words)

        self.board.params["ext_dac"]["cal"]["value"] = value



import logging

import numpy as np

from naludaq.helpers import FancyIterator

from .default import PedestalsGenerator

LOGGER = logging.getLogger("naludaq.pedestals_generator_hdsoc")


class PedestalsGeneratorHdsoc(PedestalsGenerator):
    """Pedestals generator for HDSoC.

    It's the same as the regular pedestals generator, but with a a special fix
    which accounts for the first window being invalid. When the first window fix
    is enabled, the first window of every event captured is thrown away, since
    the first window is usually (always?) invalid. This will not affect the structure
    of the generated pedestals; all windows are still present in the output.
    """

    def __init__(
        self, board, num_captures: int, num_warmup_events: int, channels: list[int]
    ):
        super().__init__(board, num_captures, num_warmup_events, channels)
        self._first_win_fix = self.board.params.get("first_window_fix", True)

    @property
    def first_window_fix(self) -> bool:
        """Get/set whether the first window fix is enabled.
        The default is True unless otherwise specified in the board parameters.

        See the class documentation for more information on the first window fix.
        """
        return self._first_win_fix

    @first_window_fix.setter
    def first_window_fix(self, enabled: bool):
        if not isinstance(enabled, bool):
            raise TypeError("Type must be boolean")
        self._first_win_fix = enabled

    def _read_window(self, start_window: int) -> tuple[int, int, int]:
        """Override for the first window fix."""
        return (
            self._correct_block_size(self.block_size),
            self._correct_start_window(start_window),
            self.board.params["windows"],
        )

    def _create_validation_stage(self, pipeline: FancyIterator) -> FancyIterator:
        """Override for the first window fix."""
        if self.first_window_fix:
            pipeline = pipeline.map(self._remove_first_window)
        return super()._create_validation_stage(pipeline)

    def _calculate_expected_window_labels(
        self, start_window: int, block_size: int
    ) -> list[int]:
        """Calculate the window labels we expect to see from an event
        read from the given block.

        Reimplmented to account for the first window fix.
        If the first window fix is enabled, the actual start window is one less
        so that we can throw it away later, and the block size is one window more
        so we can still read until the end of the block.

        Args:
            start_window (int): the start window of the block
            block_size (int): the block size

        Returns:
            list[int]: the expected window labels
        """
        return super()._calculate_expected_window_labels(
            start_window,
            self._correct_block_size(block_size),
        )[1:]

    def _remove_first_window(self, event: dict) -> dict:
        """Remove the first window from the event in-place.

        Args:
            event (dict): the event

        Returns:
            dict: the same event object
        """
        samples = self.board.params.get("samples", 32)

        event["window_labels"] = np.delete(event["window_labels"], 0, 1)
        event["data"] = np.delete(event["data"], range(samples), 1)
        event["time"] = np.delete(event["time"], range(samples), 1)
        return event

    def _correct_start_window(self, start_window: int) -> int:
        """Correct the given start window by subtracting one
        when the first window fix is enabled. Does nothing
        if the switch is off.
        """
        if self.first_window_fix:
            start_window -= 1
            if start_window < 0:
                start_window = self.board.params["windows"] - 1
        return start_window

    def _correct_block_size(self, block_size: int) -> int:
        """Correct the block size by adding one when the first
        window fix is enabled. Does nothing when the switch is off.
        """
        if self.first_window_fix:
            block_size = min(block_size + 1, self.board.params["windows"])
        return block_size

    def _validate_event(self, event: dict, expected_window_labels: list[int]) -> bool:
        """Reimplemented to check all channels.

        Windows can be missing from individual channels on HDSoC.
        """
        for channel in self.channels:
            labels = event["window_labels"][channel]
            try:
                is_bad = np.any(labels != expected_window_labels)
            except Exception as error_msg:
                LOGGER.error(f"Event validation failed due to: {error_msg}")
                return False
            if is_bad:
                LOGGER.warning(
                    "Expected windows: %s != returned: %s",
                    expected_window_labels,
                    labels,
                )
                return False
        return True



"""Worker for reading out FT60x data.

"""
import logging
import time

from naludaq.daq.workers.worker_serial_reader import SerialReader

LOGGER = logging.getLogger("naludaq.usb_reader")


class UsbReader(SerialReader):
    def __init__(self, connection, serial_buffer, *args, frequency=1000, **kwargs):
        super().__init__(connection, serial_buffer, *args, **kwargs)
        self._frequency = frequency
        self._max_interval = 0.01
        self._min_interval = 0.0001
        self._interval = self._clamp_interval(1 / frequency)

    def _reset_frequency(self):
        self._interval = 1.0 / self._frequency

    def read_serial_worker(self):
        """Worker takes raw serial data from driver buffer and stores in internal buffer.

        This solves issues with buffer overruns on higher UART baud rates.
        Runs with the polling frequency set, if to high it draws a lot of CPU power.
        If too low you get buffer overflow.

        Returns:
            appends finished packages to the package_buffer.

        """
        # Before starting to read an event, flush the buffer from whatever.
        self._reset_input_buffer()
        try:
            while self._running and self._ctrl.is_open:
                data = self._read_serial_data()
                if len(data) != 0:
                    self._store_serial_data(data)
        except KeyboardInterrupt:
            pass

    def _read_serial_data(self):
        data = None
        try:
            # read all that is there or wait for one byte (blocking)
            data = self._ctrl.read(self._ctrl.in_waiting or 2)
        except Exception as error_msg:
            # probably some I/O problem such as disconnected USB serial
            LOGGER.error(error_msg)
            self._running = False
            data = None
        if len(data) <= 0:
            time.sleep(self._interval)
            self._interval = self._clamp_interval(self._interval * 1.2)
        else:
            self._interval = self._clamp_interval(1.0 / self._frequency)
        return data

    def _clamp_interval(self, interval: float) -> float:
        """Clamp the given interval between `self._min_interval` and `self._max_interval`."""
        return min(max(interval, self._min_interval), self._max_interval)



"""Class for saving/loading acquisitions, events, and pedestals
in HDF5 format. See the README for more information on the data format.

Events are grouped into sets of fixed size called "blocks." The
events within a block are stitched together into large matrices.
Storing the events in blocks is necessary because for extremely
large events, a single matrix is too much for HDF5 to handle.
Additionally, this opens up the possibility for writing data to disk
as soon as it comes back from the board.
"""
import itertools
import logging
import math
import os
from collections import deque
from datetime import datetime
from pathlib import Path
from typing import Deque, List, Union

import h5py
import numpy as np

import naludaq
from naludaq.board import Board
from naludaq.board.params import _calculate_channel_shift
from naludaq.helpers.exceptions import BadDataError
from naludaq.models import Acquisition, acq_converters
from naludaq.parsers import get_parser

LOGGER = logging.getLogger(__name__)


# Used to load values as the correct type
_DATETIME_CONVERTER = lambda x: datetime.fromtimestamp(x)
_PATHLIB_CONVERTER = lambda x: Path(x)
_BYTES_CONVERTER = lambda x: x[()].tobytes()
_INT_CONVERTER = lambda x: int(x)


class HDF5File:
    def __init__(
        self, path: Path, mode: str, board: Board, compression="gzip", block_size=1024
    ) -> None:
        """IO class for reading and writing NaluDAQ data structures
        to/from an HDF5 file.

        # WARNING
        This class cannot handle events from boards other than specified in the
        `board` argument.

        Args:
            path (Path): the path of the file to read/write
            mode (str): the mode, either 'r' (read) or 'w' (write)
            board (Board): the board object. If mode='r', then can be None.
            compression (str): the compression format to use when writing datasets.
                Must be 'gzip' or None.
            block_size (int): the maximum number of events that can fit inside a block.

        Raises:
            PermissionError if the file cannot be opened
            FileNotFoundError if reading and the file does not exist
            BaseException in any other cases (TODO: determine cases)
        """
        if type(mode) != str:
            raise TypeError(f"Mode must be a string, not {type(mode)}")
        elif mode.lower() not in ["r", "w"]:
            raise ValueError(f"Mode must be either 'r' or 'w', not {mode}")
        elif not isinstance(path, (str, Path)):
            raise TypeError(f"Path must be str or Path, not {type(path)}")
        elif type(board) != Board and mode.lower() == "w":
            raise TypeError(f"Board must be a naludaq.board.Board, not {type(board)}")
        elif compression is not None and type(compression) != str:
            raise TypeError(f"Compression must be a string, not {type(compression)}")
        elif compression not in ["gzip", None]:
            raise ValueError(f"Compression must be 'gzip' or None, not {compression}")
        elif type(block_size) != int:
            raise TypeError(f"Block size must be an int, not {type(block_size)}")
        elif block_size <= 0:
            raise ValueError(f"Block size must be positive")

        self.path = Path(path)
        self._board = board
        self._cancel = False
        self._progress = []
        self._mode = mode
        self._compression = compression
        self._block_size = block_size

        try:
            self._file = h5py.File(str(self.path), self._mode)
        except PermissionError as e:
            LOGGER.error(f"Cannot open file: {e}")
            raise e
        except FileNotFoundError as e:
            LOGGER.error(f"File does not exist: {e}")
            raise e
        except BaseException as e:
            LOGGER.error(f"Failed to import HDF5 acquisition: {e}")
            raise e

    @property
    def has_events(self) -> bool:
        """Gets whether the file contains events."""
        return "events" in self._file

    @property
    def has_metadata(self) -> bool:
        """Gets whether the file contains acquisition metadata."""
        return "metadata" in self._file

    @property
    def has_pedestals(self) -> bool:
        """Gets whether the file contains pedestals calibration data."""
        return "pedestals" in self._file.get("calibration", {})

    @property
    def has_caldata(self) -> bool:
        """Gets whether the file contains adc2mv calibration data."""
        return "caldata" in self._file.get("calibration", {})

    @property
    def has_timingcal(self) -> bool:
        """Gets whether the file contains timing calibration data."""
        return "timingcal" in self._file.get("calibration", {})

    @property
    def num_events(self) -> int:
        """Gets the number of events the file contains."""
        if self.has_events:
            return self._file["events"].attrs["event_count"]
        else:
            return 0

    @property
    def num_blocks(self) -> int:
        """Gets the number of blocks the file contains."""
        if self.has_events:
            return len(self._file["/events/blocks"])
        return 0

    @property
    def block_size(self) -> int:
        """Gets the size of event blocks.

        If mode is 'r' and the acquisition contains events,
        the block size specification contained in the file is returned.
        Otherwise, the value specified in this class' constructor
        is returned.
        """
        if self._mode == "r" and self.has_events:
            return self._file["events"].attrs["max_block_size"]
        return self._block_size

    @property
    def enabled_channels(self):
        """Gets a list of enabled channels stored in the file.
        Returns None if the file is not an acquisition.
        """
        if self.has_events:
            return list(self._file["events"].attrs["enabled_channels"][()])
        return None

    @property
    def num_channels(self):
        """Gets the total number of channels from the file.
        Returns 0 if the file is not an acquisition.
        """
        if self.has_events:
            return self._file["events"].attrs["channel_count"]
        elif self.has_metadata:
            return self._file["metadata"]["settings"]["params"].attrs["channels"]
        return 0

    @property
    def progress(self):
        """Get/set the progress list. In a multithreaded context this can be used to read the
        status of a long-running function.

        When setting the progress list, the object must have an `append` method.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, _, __):
        """Exit method for context manager. Closes the file.

        If the `cancel()` method was called or an exception was raised
        while writing, the partial output file is removed.
        """
        try:
            self._file.close()
        except:
            pass

        # Remove partial file if erred or canceled
        if self._mode == "w" and (exc_type or self._cancel):
            try:
                os.remove(self.path)
            except:
                pass

        self.progress.append((100, "Done"))

    def cancel(self):
        """Cancels any and all IO operations as soon as possible.
        If writing, the output file is deleted.
        If reading, no result is returned.

        Can only be called from a separate thread.
        """
        self._cancel = True

    # ============================= Write Operations =============================
    def write_acquisition(
        self,
        acq: Union[deque, list, Acquisition],
        calibration: bool = True,
        metadata: Union[bool, Acquisition] = True,
    ):
        """Writes an acquisition to the file.

        # WARNING
        All events given must be from the _same_ acquisition, or this function
        may fail.

        Args:
            acq (deque, list, or Acquisition): the acquisition to write.
            calibration (bool, dict): whether to include pedestals in the output file.
            metadata (bool, Acquisition): whether to include acquisition metadata
                in the output file. If `metadata` is an Acquisition object,
                it is used as the source of metadata for the output file -- this is
                useful if `acq` is a subset of an original acquisition.

        Raises:
            TypeError if one or more of the arguments are of an invalid type.
            ValueError if the acquisition or pedestals are in an invalid format.
        """
        if not isinstance(acq, (deque, list, Acquisition)):
            raise TypeError(
                f"'acq' must be a deque, list, or Acquisition, not {type(acq)}"
            )
        elif not isinstance(calibration, bool):
            raise TypeError(
                f"'calibration' must be a bool or dict, not {type(calibration)}'"
            )
        elif not isinstance(metadata, (bool, Acquisition)):
            raise TypeError(
                f"'metadata' must be a bool or Acquisition, not {type(metadata)}"
            )
        elif len(acq) == 0:
            raise ValueError("Must have at least one event!")

        self.write_events(acq)

        if calibration and not self._cancel:
            self.progress.append((95, "Writing calibration"))
            self.write_calibration(acq)

        if metadata and not self._cancel:
            self.progress.append((99, "Writing acquisition metadata"))
            self.write_metadata(acq)

    def write_events(self, events: Union[Deque[dict], List[dict]]):
        """Writes a bunch of events to the file.

        Data is stored into the 'events' group as a handful of datasets.
        Each dataset stores a particular attribute of _all_ events,
        and are indexed by the event number.

        Channels that are disabled (i.e. empty in the event data) are _not_
        written to the datasets in order to save space. The channel number
        corresponding to an axis in the datasets can be determined by the
        'enabled_channels` attribute.

        Args:
            events (deque, list): a deque or list of events.

        Raises:
            ValueError if there is not at least one event.
            BadDataError if there are no valid events in the acquisition
        """
        if not events:
            raise ValueError("Acquisition must have at least one event.")

        # Look ahead to the first valid event to get shape characteristics since
        # self._board cannot be trusted to have the same model/settings as the acquisition
        try:
            attrs = self._get_acq_attributes(events)
            enabled_channels = attrs["enabled_channels"]
            channel_count = attrs["channel_count"]
        except BadDataError:
            raise

        # Block size specified by user
        num_events = len(events)
        num_blocks = math.ceil(num_events / self._block_size)

        events_group = self._file.create_group("events")
        events_group.attrs["max_block_size"] = self._block_size
        events_group.attrs["event_count"] = num_events
        events_group.attrs.create(
            "enabled_channels", np.array(enabled_channels), dtype=np.uint16
        )
        events_group.attrs.create("channel_count", data=channel_count)

        block_links = events_group.create_dataset(
            "blocks", shape=(num_blocks,), dtype=h5py.ref_dtype
        )

        # Split events into blocks and write
        for block_idx in range(num_blocks):
            LOGGER.info(f"Writing block ({block_idx+1}/{num_blocks})")
            self.progress.append(
                (
                    85 * block_idx / num_blocks,
                    f"Writing block ({block_idx+1}/{num_blocks})",
                )
            )

            if block_idx == num_blocks - 1:
                block_events = list(
                    itertools.islice(events, block_idx * self._block_size, num_events)
                )
            else:
                block_events = list(
                    itertools.islice(
                        events,
                        block_idx * self._block_size,
                        (block_idx + 1) * self._block_size,
                    )
                )

            block_group = events_group.create_group(f"block_{block_idx}")
            block_links[block_idx] = block_group.ref
            self._write_event_block(block_group, block_events, attrs)
            if self._cancel:
                return

    def _write_event_block(self, group: h5py.Group, events: list, attrs: dict):
        """Writes a list of events to a block group.

        Args:
            parser (Parser): the parser to use for unprocessed events
            group (h5py.Group): the block group
            events (list): the events to write
            attrs (dict): the acquisition attributes from `_get_acq_attributes()`
        """
        parser = attrs["parser"]
        num_events = len(events)

        window_label_shape = attrs["window_labels"]
        channels, samples = attrs["event_data"]
        enabled_channels = attrs["enabled_channels"]

        # Allocate a bunch of empty arrays, one for each type of event property
        data = np.full((num_events, channels, samples), fill_value=0, dtype=np.uint16)
        window_labels = np.full(
            (num_events, *window_label_shape), fill_value=0, dtype=np.uint16
        )
        time = np.full((num_events, channels, samples), fill_value=0, dtype=np.uint16)
        pkg_nums = np.empty((num_events,), dtype=np.uint32)
        event_nums = np.empty((num_events,), dtype=np.uint32)
        start_windows = np.empty((num_events,), dtype=np.uint32)
        creation_times = np.empty((num_events,), dtype=np.uint64)
        bad_events = []
        event_names = {}  # Mapping of event_idx: name

        # Load up the arrays before writing to disk. One row in the arrays corresponds to one event
        for idx, event in enumerate(events):

            # Some stuff needs to be stored before parsing, in case the event is bad
            pkg_nums[idx] = event.get("pkg_num", 0)
            event_nums[idx] = event.get("event_num", 0)

            # Storing names for all events is wasteful; only store user-defined names
            name = event.get("name", None)
            if name:
                event_names[idx] = name.encode("ascii")

            # HDF5 acquisitions need parsed data
            if "data" not in event:
                try:
                    parsed = parser.parse(event)
                    parsed["name"] = event.get("name", None)
                except BadDataError as e:
                    LOGGER.info(f"Event {idx} not saved due to parsing error: {e}")
                    bad_events.append(idx)
                    continue
            else:
                parsed = event

            # Some channels may be disabled, need to store channel-by-channel.
            for chan_idx, chan in enumerate(enabled_channels):
                data[idx, chan_idx, :] = np.array(parsed["data"][chan])
                window_labels[idx, chan_idx, :] = np.array(
                    parsed["window_labels"][chan]
                )
                # if idx <= 10:
                # print(parsed['time'][chan])
                time[idx, chan_idx, :] = np.array(parsed["time"][chan], dtype=np.uint16)

            start_windows[idx] = parsed["start_window"]
            creation_times[idx] = parsed["created_at"]

            if self._cancel:
                return

        # Write all the arrays and whatnot to the file
        # self.progress.append((90, 'Storing events'))
        group.create_dataset("package_numbers", data=pkg_nums, dtype=np.uint32)
        group.create_dataset("event_numbers", data=event_nums, dtype=np.uint32)
        group.create_dataset("start_windows", data=start_windows, dtype=np.uint16)
        group.create_dataset("creation_times", data=creation_times, dtype=np.uint64)
        group.create_dataset(
            "window_labels",
            data=window_labels,
            dtype=np.uint16,
            compression=self._compression,
            compression_opts=9,
        )
        group.create_dataset(
            "data",
            data=data,
            dtype=np.uint16,
            compression=self._compression,
            compression_opts=9,
        )
        group.create_dataset(
            "time",
            data=time,
            dtype=np.uint16,
            compression=self._compression,
            compression_opts=9,
        )
        if bad_events:
            dataset = np.array(bad_events, dtype=np.uint)
            group.attrs.create("bad_events", data=dataset, dtype=np.uint)
        if event_names:
            # A list of ([event #, event name]), ends up being saved as an array of [bytes, bytes]
            event_names = np.array(list(event_names.items()))
            group.attrs.create(
                "names",
                shape=(len(event_names), 2),
                data=event_names,
                dtype=event_names.dtype,
            )

    def _get_acq_attributes(self, events: Union[Deque[dict], List[dict], Acquisition]):
        """Finds the shape of window labels/event data datasets and the enabled channels
        based on the first good event in the acquisition.

        Args:
            events (deque, list, Acquisition): the acquisition

        Raises:
            BadDataError if there are no valid events

        Returns:
            A dict of {
                'window_labels': (channel count, window count),
                'event_data': (channel count, sample count),
                'enabled_channels': [0, 1, 2, 3, 4, ...]
            }
        """
        parser = get_parser(_get_params(events, self._board))

        # Look ahead to the first valid event to get shape characteristics since
        # self._board cannot be trusted to have the same model/settings as the acquisition
        for idx, event in enumerate(events):
            try:
                # We're really only parsing one good event, so storing the output has negligible benefits
                parsed = event if "data" in event else parser.parse(event)

                samples = max([len(x) for x in parsed["data"]])
                channels = len(parsed["data"])
                enabled_channels = [
                    x for x in range(channels) if len(parsed["data"][x]) != 0
                ]
                label_count = max([len(x) for x in parsed["window_labels"]])

                # We only need valid data from one event
                return {
                    "window_labels": (len(enabled_channels), label_count),
                    "event_data": (len(enabled_channels), samples),
                    "enabled_channels": enabled_channels,
                    "channel_count": channels,  # total number of channels
                    "parser": parser,
                }
            except BadDataError as e:
                LOGGER.warning(f"Event {idx} not saved due to parser error: {e}")
            except BaseException as e:
                LOGGER.warning(f"Event {idx} not saved due to: {e}")

        # No way we can continue if we can't determine the shape
        raise BadDataError("No valid events found in acquisition")

    def write_calibration(self, acq: Acquisition):
        """Write all calibration to the file.

        If the acquisition is missing one or more types of calibration,
        there is no corresponding entry in the output file.

        ## Currently only supports pedestals.

        Args:
            acq (Acquisition): the acquisition to take calibrations from
        """
        self._file.require_group("calibration")
        if acq.pedestals:
            self.write_pedestals(acq.pedestals)
        # if acq.caldata: self.write_caldata() <---- not yet supported
        # if acq.adc2mv: self.write_adc2mv() <--- not yet supported

    def write_pedestals(self, pedestals: dict):
        """Writes a pedestals dict to the file.
        Data is stored into the 'calibration/pedestals' group.

        Args:
            pedestals (dict): the pedestals dict.

        Raises:
            ValueError if the pedestals dict is invalid
        """
        if not pedestals:
            raise ValueError("No pedestals to write")

        pedestals_data = np.array(pedestals["data"], dtype=np.uint16)

        group = self._file.require_group("calibration/pedestals")
        group.attrs["model"] = self._board.model
        group.attrs["channels"] = self._board.channels
        group.attrs["samples"] = self._board.params.get(
            "samples", len(pedestals["data"][0][0])
        )
        group.attrs["windows"] = self._board.params.get(
            "windows", len(pedestals["data"][0])
        )
        group.create_dataset(
            "data", data=pedestals_data, dtype=np.uint16, compression=self._compression
        )

    def write_event(self, event: dict):
        """Writes a single event to the file.
        This internally uses the `write_acquisition` function for
        format consistency.

        Args:
            event (dict): the event dict.
        """
        self.write_acquisition([event], calibration=False, metadata=False)

    def write_metadata(self, acq: Acquisition):
        """Stores metadata from an acquisition into a group.

        Adds several attributes to the given group.
        If `acq` is an Acquisition object, then an 'info' group is created
        holding the acquisition metadata (e.g. info, trigger_settings).

        Args:
            acq (deque, list, Acquisition): the acquisition to steal metadata from
        """
        group = self._file.require_group("metadata")
        _store_attributes(group, acq.as_dict()["metadata"])

    # ============================= Read Operations =============================
    def read_acquisition(self) -> Acquisition:
        """Reads an acquisition from the file.

        Returns:
            An acquisition object, or None if the operation
            was canceled.

        Raises:
            AttributeError if data is missing from the file.
        """
        try:
            events = self.read_events()
            if self._cancel:
                return None

            data = {
                "events": events,
                "metadata": self.read_metadata() if self.has_metadata else None,
                "calibration": {
                    "calibration": None,  # Not yet supported
                    "pedestals": self.read_pedestals() if self.has_pedestals else None,
                    "timingcal": None,  # Not yet supported
                },
            }
            acq = acq_converters.upgrade_old_acquisition(data)
        except AttributeError:
            raise

        if self._cancel:
            return None

        return acq

    def read_events(self, buffer: Union[deque, list, Acquisition] = None):
        """Reads all events from the file.

        Args:
            buffer (deque): the buffer to read into. If no buffer is
                provided, a new deque is used instead.

        Raises:
            AttributeError if one or more attributes does not exist
                in the 'events' group, or the group does not exist.

        Returns:
            The buffer, or a new deque if no buffer was provided, or
            None if the operation was canceled.
        """
        if buffer is None:
            buffer = deque()

        group = self._file.get("events", None)
        if not group:
            raise AttributeError("File does not contain events")

        # Pull all the datasets from the file
        try:
            num_events = group.attrs["event_count"]
        except AttributeError as e:
            LOGGER.error(f"Malformed data, failed to read attribute: {e}")
            raise

        # Allocate space in buffer ahead of time
        offset = len(buffer)
        buffer.extend([None] * num_events)

        # `blocks` dataset contains references that point to block groups
        block_refs = self._file["/events/blocks"]
        block_size = self.block_size
        num_blocks = self.num_blocks

        # Load up the output buffer block-by-block
        for block_idx in range(num_blocks):
            self.progress.append(
                (
                    block_idx / num_blocks * 100,
                    f"Reading block {block_idx+1}/{num_blocks}",
                )
            )
            block_group = self._file[block_refs[block_idx]]
            self._read_event_block(block_group, buffer, offset + block_idx * block_size)

            if self._cancel:
                return None

        return buffer

    def _read_event_block(
        self, group: h5py.Group, buffer: Union[deque, list, Acquisition], offset: int
    ):
        """Reads events from a block into a buffer.

        Args:
            group (int): the group to read from
            buffer (deque): the output buffer for events
            offset (int): the offset in the buffer to start at
        """
        # Pull all the datasets from the file
        try:
            data = group["data"]
            time = group["time"]
            window_labels = group["window_labels"]
            pkg_nums = group["package_numbers"]
            event_nums = group["event_numbers"]
            start_windows = group["start_windows"]
            creation_times = group["creation_times"]
            bad_events = group.attrs.get("bad_events", [])
            names = group.attrs.get("names", [])

            num_events = len(data)
            enabled_channels = self.enabled_channels
            channel_count = self.num_channels
        except AttributeError as e:
            LOGGER.error(f"Malformed data, failed to read attribute: {e}")
            raise

        # Create an event dict for each row in the arrays
        for event_idx, buffer_idx in enumerate(range(offset, offset + num_events)):
            if event_idx in bad_events:
                event = {
                    "rawdata": b"",  # raw data isn't stored in HDF5
                    "pkg_num": pkg_nums[event_idx][()],
                    "event_num": event_nums[event_idx][()],
                }
                buffer[buffer_idx] = event
                LOGGER.debug(f"Found bad event #{event_idx}")
                continue

            event = {
                "data": [[] for _ in range(channel_count)],
                "time": [[] for _ in range(channel_count)],
                "window_labels": [[] for _ in range(channel_count)],
                "pkg_num": pkg_nums[event_idx][()],
                "event_num": event_nums[event_idx][()],
                "start_window": start_windows[event_idx],
                "created_at": creation_times[event_idx],
            }
            buffer[buffer_idx] = event

            # Some channels are disabled, need to look at enabled_channels to map into dict
            for chan_idx, chan in enumerate(enabled_channels):
                event["data"][chan] = data[event_idx, chan_idx, :][()]
                event["time"][chan] = time[event_idx, chan_idx, :][()]
                event["window_labels"][chan] = window_labels[event_idx, chan_idx, :][()]

            # Some channels are disabled, need to look at enabled_channels to map into dict
            for chan_idx, chan in enumerate(enabled_channels):
                event["data"][chan] = data[event_idx, chan_idx, :][()]
                event["time"][chan] = time[event_idx, chan_idx, :][()]
                event["window_labels"][chan] = window_labels[event_idx, chan_idx, :][()]

            # Some channels are disabled, need to look at enabled_channels to map into dict
            for chan_idx, chan in enumerate(enabled_channels):
                event["data"][chan] = data[event_idx, chan_idx, :][()]
                event["time"][chan] = time[event_idx, chan_idx, :][()]
                event["window_labels"][chan] = window_labels[event_idx, chan_idx, :][()]

            if self._cancel:
                return None

        # Unnecessary to loop over the entire list for each event, so do it after
        for entry in names:
            event_num = int(entry[0].decode("ascii"))
            buffer[event_num + offset]["name"] = entry[1].decode("ascii")

            if self._cancel:
                return None

    def read_pedestals(self) -> dict:
        """Reads pedestals from the file.

        Returns:
            The pedestals dict, or None if the operation
            was canceled.

        Raises:
            AttributeError if the file does not have pedestals,
                or they are stored in an invalid format.
        """
        group = self._file.get("calibration/pedestals", None)
        if not group:
            raise AttributeError("File does not contain pedestals")

        data = group.get("data", None)
        if not data:
            raise AttributeError("Pedestals are stored in an invalid format")

        peds = {
            "data": data[()],
        }
        if self._cancel:
            return None

        return peds

    def read_metadata(self) -> dict:
        """Reads acquisition metadata from the file.

        Returns:
            A dictionary containing the metadata, or None
            if the operation was canceled.
        """
        metadata_group = self._file.get("metadata", None)
        if metadata_group is None:
            return

        type_converters = {
            "date": _DATETIME_CONVERTER,
            "acq_start": _DATETIME_CONVERTER,
            "acq_stop": _DATETIME_CONVERTER,
            "clock_file": _PATHLIB_CONVERTER,
            "stop_word": _BYTES_CONVERTER,
            "acq_num": _INT_CONVERTER,
        }
        metadata = _load_dict(metadata_group, type_converters, recursive=True)
        acq_version = metadata.get(
            "file_version", "0.2"
        )  # 0.2 is earliest HDF5 version
        acq_converters.convert(
            metadata, acq_version, acq_converters.LATEST_VERSION, in_place=True
        )

        # TRBHM needs a string stop_word instead of bytes
        if metadata.get("model", None) in ["trbhm", "dsa-c10-8"]:
            metadata["settings"]["params"]["stop_word"] = metadata["settings"][
                "params"
            ]["stop_word"].hex()

        if self._cancel:
            return None

        return metadata


def _store_attributes(group: h5py.Group, attrs: dict, recursive=True):
    """Adds a dict of attributes to a h5py group.

    Skips entries that have 'None' value.

    Args:
        group (Group): the group to add the attributes to
        attrs (dict): a dict of {name: value}
        recursive (bool): whether to create new groups for
            entries with dict values
    """
    for k, v in attrs.items():
        k = str(k)
        if v is None:  # HDF5 doesn't support None value
            continue
        elif isinstance(v, dict) and recursive:
            new_group = group.create_group(k, track_order=True)
            _store_attributes(new_group, v)
        else:
            v, dtype = _convert_for_storing(v)
            group.attrs.create(k, v, dtype=dtype)


def _load_dict(
    group: h5py.Group,
    type_conversions={},
    load_datasets=True,
    load_attrs=True,
    recursive=True,
) -> dict:
    """Loads a simple dictionary from objects in a group.

    Note that the type conversions are applied to ALL dict
    entries with the matching name, regardless of whether
    it is nested in another dict.

    Args:
        group (Group): the group to convert to a dict
        type_conversions (dict): a dict of {name: callable},
            applied to entries that have the same name. Useful
            for converting to some types like datetimes. The
            callable is given one argument: the object in
            its default h5py type.
        load_datasets (bool): whether to load datasets
        load_attrs (bool): whether to load attributes
        recursive (bool): whether to recursively load subgroups

    Returns:
        A dict containing translated data from the group.
    """
    result = {}
    type_conversions = type_conversions or {}

    if recursive:
        for k, v in group.items():
            if isinstance(v, h5py.Group):
                result[k] = _load_dict(
                    v, type_conversions, load_datasets, load_attrs, recursive
                )

    if load_datasets:
        for k, v in group.items():
            if isinstance(v, h5py.Group):
                continue
            elif (
                k in type_conversions
            ):  # TODO: add support for differentiating converters for nested dicts
                result[k] = type_conversions[k](v)
            else:
                result[k] = v[()]

    if load_attrs:
        for k, v in group.attrs.items():
            if (
                k in type_conversions
            ):  # TODO: add support for differentiating converters for nested dicts
                v = type_conversions[k](v)
            elif not isinstance(v, str):
                v = v[()]

            result[k] = v

    return result


def _convert_for_storing(value):
    """Converts a value to a type that can be stored in HDF5

    datetime -> float
    PathLike -> str
    bytes -> np.void

    Args:
        value: the value to convert

    Returns:
        A tuple of (converted value, new dtype)
    """
    dtype = None
    if isinstance(value, datetime):
        value = value.timestamp()
    elif isinstance(value, os.PathLike):
        value = str(value)
    elif isinstance(value, bytes):
        value = np.void(value)
    return value, dtype


def _get_package_versions():
    """Gets the NaluDAQ and NaluScope version strings.

    If the naluscope version cannot be determined (e.g. if it is not
    installed), then the version is returned as None

    Returns:
        A tuple of (NaluDAQ version, NaluScope version)
    """
    naludaq_version = naludaq.__version__
    try:
        import naluscope

        naluscope_version = naluscope.__version__
    except:
        naluscope_version = None

    return naludaq_version, naluscope_version


def _calculate_event_progress(num_events, event_num, min=10, max=90):
    """Calculates the progress percentage for operations that
    iterate over many events (e.g. loading, saving).

    This function can also be used to limit the number of
    progress updates emitted to avoid flooding the progress receiver:
    this function returns None if a progress update should not be
    emitted. Limits progress updates to 1 per percentage point.

    Args:
        num_events (int): the total number of events
        event_num (int): the current event number
        min (int): the starting percentage
        max (int): the ending percentage

    Returns:
        The percentage, or None if a progress update should not be emitted.
    """
    notification_interval = num_events / (max - min) or 1

    if event_num % notification_interval < 1:
        return min + event_num / notification_interval
    else:
        return None


def _get_params(acq: Acquisition, board) -> dict:
    """Get params based on info in acquisition object.

    If acquisition object doesn't contain params try using the model from the acquisition
    if model is not available, return params from the board.
    """
    if isinstance(acq, Acquisition) and acq.metadata:
        params = acq.params
        if params.get("chanshift", None) is None:
            params["chanshift"] = _calculate_channel_shift(params["chanmask"])
        return acq.params
    return board.params



import logging
import time
from typing import Iterable, Callable

from naludaq.controllers import get_board_controller, get_readout_controller
from naludaq.helpers import FancyIterator, validations
from naludaq.helpers.exceptions import BadDataError
from naludaq.helpers.helper_functions import event_transfer_time
from naludaq.tools.data_collector._daq_interface import get_daq_interface

logger = logging.getLogger("naludaq.data_collector")


class DataCollector:
    def __init__(
        self,
        board,
        channels: "list[int] | None" = None,
        trig: "str|None" = None,
        lb: "str|None" = None,
        acq: "str|None" = None,
        ped: "str|None" = None,
        readoutEn: "bool|None" = None,
        singleEv: "bool|None" = None,
        windows: "int|None" = None,
        lookback: "int|None" = None,
        write_after_trig: "int|None" = None,
    ):
        """Tool for easily collecting data from a board.

        This tool implements a simple interface for collecting data from a board
        without having to worry about the details of the readout configuration and DAQ.
        This tool also provides stability and robustness over a hand-wrangled readout
        solution.

        If the data collection succeeds, the following guarantees are made:
        - The events are all parsed and valid
        - The number of events is exactly the number of events requested
        - The events are in chronological order
        - The events are not duplicated

        Note that this tool does not preserve the state of the board before and after
        data collection. It is the responsibility of the user to ensure that the board
        state is restored if necessary.

        A DataCollector object may be reused as many times as desired, as long as only
        one collection is running at a time. The benefit of reusing a DataCollector
        is that the readout does not need to be reconfigured each time, and the
        timeout starts off better-tuned on subsequent collections.

        The data collector is modeled after Rust's iterators. Operations may be chained
        sequentially to build up a pipeline of operations to perform on the data. The
        pipeline is executed when the final `collect()` function is called.

        Example usage:

        .. code-block:: python

            collector = DataCollector(board)
            collector
                .iter(attempts=3)                 # 1. create a new iterator
                .enumerate()                      # 2. add an index to each event
                .filter(lambda x: x[0] % 2 == 0)  # 3. filter out odd-numbered events
                .unenumerate()                    # 4. remove the index so we're left with events only
                .take(10)                         # 5. only take the first 10 odd-numbered events (1, 3, .., 21)
                .collect()                        # 6. execute the pipeline

        Properties:
            board: Board object
            readout_settings: dictionary of readout settings to use
            window_settings: the window of the sampling array to read
            forced: whether the readout is in forced mode or not
            channels: list of channels to read

        Args:
            board: Board object
            channels (Iterable[int]): list of channels to read, defaults to all channels
            trig (str): trigger mode, defaults to "ext"
            lb (str): lookback mode, defaults to "trigrel"
            acq (str): acquisition mode, defaults to "raw"
            ped (str): pedestal mode, defaults to "zero"
            readoutEn (bool): whether to enable readout, defaults to True
            singleEv (bool): whether to enable single event readout, defaults to False
            windows (int): number of windows to read, defaults to 8
            lookback (int): number of lookback windows, defaults to 8
            write_after_trig (int): number of windows to write after trigger, defaults to 8
        """
        self._board = board
        self.channels = channels or range(board.channels)
        self.readout_settings = {
            "trig": trig or "ext",
            "lb": lb or "trigrel",
            "acq": acq or "raw",
            "ped": ped or "zero",
            "readoutEn": readoutEn or True,
            "singleEv": singleEv or False,
        }
        self._window_settings = {
            "windows": 8,
            "lookback": 8,
            "write_after_trig": 8,
        }
        self.set_window(windows, lookback, write_after_trig)
        self.trigger_fn = None
        self._single_event_timeout = 0
        self._margin = 2
        self._trigger_limit = 0.01
        self._last_trigger_time = None
        self.timeout_overhead = self.board.params.get("timeout_overhead", 10.0)

    @property
    def board(self):
        """Get the board object used by the data collector"""
        return self._board

    @property
    def readout_settings(self) -> dict:
        """Get/set the readout settings dictionary.

        This dictionary contains the following keys, which correspond to the
        start_readout() function in the board controller:
            "trig"
            "lb"
            "acq"
            "ped"
            "readoutEn"
            "singleEv"
        """
        return self._readout_settings

    @readout_settings.setter
    def readout_settings(self, settings: dict):
        validations.validate_readout_settings(settings)
        self._readout_settings = settings.copy()

    @property
    def window_settings(self) -> dict:
        """Get the window settings dictionary.

        This dictionary contains the following keys:
            "windows"
            "lookback"
            "write_after_trig"
        """
        return self._window_settings

    @property
    def channels(self) -> list[int]:
        """Get/set the channels to read"""
        return self._channels

    @channels.setter
    def channels(self, channels: Iterable[int]):
        validations.validate_channel_sequence_or_raise(self.board.params, channels)
        self._channels = list(channels)

    @property
    def forced(self) -> bool:
        """Get/set whether the readout is in forced mode or not"""
        return self._readout_settings["lb"].lower().startswith("f")

    @forced.setter
    def forced(self, forced: bool):
        if not isinstance(forced, bool):
            raise TypeError("Value must be a boolean")
        if forced:
            self._readout_settings["lb"] = "forced"
        else:
            self._readout_settings["lb"] = "trigrel"

    @property
    def trigger_interval_limit(self) -> float:
        """Get/set the trigger interval limit in seconds.

        This is the minimum time between consecutive triggers.
        Note that this increases the calculated event timeout
        by the same amount to account for dead time.

        Some boards don't like it when you send triggers really
        fast, so this is a way to prevent that.
        """
        return self._trigger_limit

    @trigger_interval_limit.setter
    def trigger_interval_limit(self, limit: float):
        if not isinstance(limit, float):
            raise TypeError("Value must be a float")
        if limit < 0:
            raise ValueError("Value must be non-negative")
        self._trigger_limit = limit

    @property
    def trigger_frequency_limit(self) -> float:
        """Same thing as self.trigger_interval_limit, but in Hz."""
        return 1 / self._trigger_limit

    @trigger_frequency_limit.setter
    def trigger_frequency_limit(self, limit: float):
        if limit == 0:
            raise ValueError("Zero frequency not allowed")
        self.trigger_interval_limit = 1 / limit

    @property
    def trigger_fn(self) -> Callable:
        """Get/set the trigger callback function.

        Settings this function allows the user to inject custom logic
        into the data capture code.
        """
        return self._trigger_fn

    @trigger_fn.setter
    def trigger_fn(self, fn: Callable):
        if fn is None:
            fn = get_board_controller(self._board).toggle_trigger
        validations.validate_callable_or_raise(fn)
        self._trigger_fn = fn

    @property
    def timeout_overhead(self) -> float:
        """Get/set the timeout offset in seconds.

        This is the amount of time added to the event timeout.
        """
        return self._overhead

    @timeout_overhead.setter
    def timeout_overhead(self, value: float):
        if value < 0:
            raise ValueError("Value must be non-negative")
        self._overhead = value
        self._recompute_timeout()

    def set_window(
        self, windows: int = None, lookback: int = None, write_after_trig: int = None
    ):
        """Set the read window.

        Args:
            windows (int): number of windows to read for each event
            lookback (int): number of windows to look back after writing "write_after_trig" windows
            write_after_trig (int): number of windows to write after the trigger
        """
        if windows is not None:
            self._window_settings["windows"] = windows
        if lookback is not None:
            self._window_settings["lookback"] = lookback
        if write_after_trig is not None:
            self._window_settings["write_after_trig"] = write_after_trig

    def set_self_trigger(self):
        """Sets trigger to self mode"""
        self._readout_settings["trig"] = "self"

    def set_external_trigger(self):
        """Sets trigger to external mode"""
        self._readout_settings["trig"] = "ext"

    def set_immediate_trigger(self):
        """Sets trigger to immediate mode"""
        self._readout_settings["trig"] = "immediate"

    def iter(self, count: int, attempts: int = 3) -> FancyIterator:
        """Create a bounded iterator of valid events.

        The iterator lazily generates events, and is a FancyIterator
        so it can be used as a data pipeline.

        It is encouraged to use this method instead of `iter_inf()` where
        possible to avoid collecting data forever.

        When any one event takes more than `attempts` attempts to capture,
        a TimeoutError is raised. The attempt counter is reset on success.

        Example:

        .. code-block:: python

            collector = DataCollector(board)
            collector
                .iter(10)
                .collect()

        Args:
            count (int): maximum number of events to read
            attempts (int): number of attempts to get an event

        Returns:
            FancyIterator: an iterator over events
        """
        return self.iter_inf(attempts).take(count)

    def iter_inf(self, attempts: int = 3) -> FancyIterator:
        """Create an infinite iterator of valid events.

        The iterator lazily generates events, and is a FancyIterator
        so it can be used as a data pipeline.

        When any one event takes more than `attempts` attempts to capture,
        a TimeoutError is raised. The attempt counter is reset on success.

        Example:

        .. code-block:: python

            collector = DataCollector(board)
            collector
                .iter_inf()
                .take(10)
                .collect()

        Args:
            attempts (int): number of attempts to get an event

        Returns:
            FancyIterator: an iterator over valid events
        """
        return FancyIterator(self._iter_inner(attempts))

    def _iter_inner(self, attempts=3):
        """A generator of single events.

        This method lazily generates events. If any one event takes more
        than `attempts` attempts to capture, a TimeoutError is raised.
        The attempt counter is reset on success.

        The timeout per event is increased on each failure to capture an event.
        This is to account for real-world factors which may make the initial
        timeout computation unrealistically short.

        Args:
            attempts (int): number of attempts to get an event
        """
        validations.validate_positive_int_or_raise(attempts)
        validations.validate_readout_settings(self._readout_settings)

        # the number of channels may have changed, need to recompute
        # (the timeout margin isn't affected by this)
        self._recompute_timeout()
        self._setup_readout()
        self._start_readout()
        self._last_trigger_time = 0
        try:
            for event in self._capture_with_attempts(attempts):
                yield event
        finally:
            self._stop_readout()

    def _capture_with_attempts(self, attempts: int) -> Iterable[dict]:
        """Attempts to capture events until the attempt counter is reached.

        This method lazily generates events. If any one event takes more
        than `attempts` attempts to capture, a TimeoutError is raised.
        The attempt counter is reset on success.

        The timeout per event is increased on each failure to capture an event.
        This is to account for real-world factors which may make the initial
        timeout computation unrealistically short.

        Raises:
            TimeoutError: if the attempt counter is reached for an event
        """
        attempt_count = 0
        interface = get_daq_interface(self._board)
        interface.start_capture()
        try:
            while True:
                self._toggle_trigger()
                try:
                    for event in interface.stream(self._single_event_timeout):
                        yield event
                        attempt_count = 0
                        self._toggle_trigger()
                except TimeoutError:
                    logger.warning("Capture failed, increasing sleep time")
                    self._event_failed()
                    attempt_count += 1
                except BadDataError:
                    logger.warning("Bad event received, ignoring")
                    attempt_count += 1
                if attempt_count >= attempts:
                    raise TimeoutError(
                        f"Failed to capture event after {attempts} attempts"
                    )
        finally:
            interface.stop_capture()

    # ================================================================================
    # Readout
    # ================================================================================
    def _start_readout(self):
        """Start a readout"""
        logger.info("Starting readout using: %s", self._readout_settings)
        get_board_controller(self._board).start_readout(**self._readout_settings)

    def _stop_readout(self):
        """Stop the readout"""
        logger.info("Stopping readout")
        get_board_controller(self._board).stop_readout()

    def _setup_readout(self):
        """Set up the readout"""
        rc = get_readout_controller(self.board)
        rc.set_readout_channels(self.channels)
        rc.set_read_window(**self._window_settings)

    def _toggle_trigger(self):
        """Toggle the trigger.

        Includes a `time.sleep` to enforce the trigger interval limit.
        """
        if self.readout_settings["trig"].lower().startswith(("i", "s")):
            return

        now = time.perf_counter()
        delta = now - self._last_trigger_time
        if delta < self._trigger_limit:
            time.sleep(self._trigger_limit - delta)
        self._last_trigger_time = now
        self._trigger_fn()

    # ================================================================================
    # Timeout
    # ================================================================================
    def _event_failed(self):
        """Callback for when an event capture times out.

        Subclasses may override this method to implement custom behavior.
        It is highly recommended to call the super method, as it handles
        increasing the timeout.
        """
        self._increase_timeout()

    def _recompute_timeout(self):
        """Recompute the event timeout attribute based on the current settings."""
        self._single_event_timeout = event_transfer_time(
            self._board,
            windows=self.window_settings["windows"],
            channels=len(self._channels),
            margin=self._margin,
            overhead=self._overhead,
        )
        logger.debug("Timeout changed to %s", self._single_event_timeout)

    def _increase_timeout(self, current_amount=0, needed_amount=1):
        """Increase the timeout based on received vs expected.

        Formula is intended to change the margin incrementally without
        causing to much oscillation.
        """
        # Settles at 0.38
        factor = ((needed_amount - current_amount) / needed_amount) / 3 + 1
        self._margin = min(self._margin * factor, 10)
        self._recompute_timeout()



from abc import ABC, abstractmethod


class SPIDevice(ABC):
    def __init__(self, daisy_chained: bool = True):
        self.daisy_chain_en = daisy_chained
        # self._chain_address

    @property
    def daisy_chain_en(self):
        return self._daisy_chained

    @daisy_chain_en.setter
    def daisy_chain_en(self, enabled: bool):
        if not isinstance(enabled, bool):
            raise TypeError("enabled must be a bool.")

        self._daisy_chained = enabled

    @abstractmethod
    def init_cmd(self):
        """Generate the initialization command for the device."""
        pass

    @abstractmethod
    def init(self):
        """Initialize the device."""
        pass

    @abstractmethod
    def _write(self, message):
        """Write a message to the device."""
        pass

    @abstractmethod
    def no_op(self):
        """Send a no-op command to the device."""
        pass
